import json

import requests


def get_summary(cid):
    url = "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/cid/%d/description/json" % cid
    result = requests.get(url)
    summary = json.loads(result.content)
    return summary


def parse_summary_for_odor(summary):
    statements = []
    # keywords should include aroma but exclude aromatic (due to its special meaning in chemistry)
    keywords = ("odor", "odour", "smell", "aroma ", "aroma,", "aroma.", "fragrance")
    if "InformationList" in summary:
        for item in summary["InformationList"]["Information"]:
            if "Description" in item:
                for statement in item["Description"].split("."):
                    if any([x in statement.lower() for x in keywords]):
                        statements.append(statement.strip())
    return statements


def get_physical_description(cid):
    url = (
        "https://pubchem.ncbi.nlm.nih.gov/rest/pug_view/data/compound/%d/JSON?heading="
        "Physical+Description"
        % cid
    )
    result = requests.get(url)
    try:
        summary = json.loads(result.content)
    except UnicodeDecodeError:
        summary = {}
    return summary


def parse_physical_description_for_odor(physical_description):
    statements = []
    try:
        strings = [
            x["Value"]["StringWithMarkup"][0]["String"]
            for x in physical_description["Record"]["Section"][0]["Section"][0]["Section"][0][
                "Information"
            ]
        ]
    except KeyError:
        pass
    else:
        # keywords should include aroma but exclude aromatic
        # (due to its special meaning in chemistry)
        keywords = ("odor", "odour", "smell", "aroma ", "aroma,", "aroma.", "fragrance")
        for string in strings:
            for statement in string.split("."):
                if any([x in statement.lower() for x in keywords]):
                    statements.append(statement.strip())
    return statements


def get_ghs_classification(cid):
    url = (
        "https://pubchem.ncbi.nlm.nih.gov/rest/pug_view/data/compound/%d/JSON?heading=GHS"
        "+Classification"
        % cid
    )
    result = requests.get(url)
    try:
        summary = json.loads(result.content)
    except UnicodeDecodeError:
        summary = {}
    return summary


GHS_CODES = {
    "H300": "Fatal if swallowed.",
    "H301": "Toxic if swallowed",
    "H302": "Harmful if swallowed",
    "H303": "May be harmful if swallowed",
    "H304": "May be fatal if swallowed and enters airways",
    "H305": "May be harmful if swallowed and enters airways",
    "H310": "Fatal in contact with skin",
    "H311": "Toxic in contact with skin",
    "H312": "Harmful in contact with skin",
    "H313": "May be harmful in contact with skin",
    "H314": "Causes severe skin burns and eye damage",
    "H315": "Causes skin irritation",
    "H316": "Causes mild skin irritation",
    "H317": "May cause an allergic skin reaction",
    "H318": "Causes serious eye damage",
    "H319": "Causes serious eye irritation",
    "H320": "Causes eye irritation",
    "H330": "Fatal if inhaled",
    "H331": "Toxic if inhaled",
    "H332": "Harmful if inhaled",
    "H333": "May be harmful if inhaled",
    "H334": "May cause allergy or asthma symptoms or breathing difficulties if inhaled",
    "H335": "May cause respiratory irritation",
    "H336": "May cause drowsiness or dizziness",
    "H340": "May cause genetic defects",
    "H341": "Suspected of causing genetic defects",
    "H350": "May cause cancer",
    "H351": "Suspected of causing cancer",
    "H360": "May damage fertility or the unborn child",
    "H361": "Suspected of damaging fertility or the unborn child",
    "H361d": "Suspected of damaging the unborn child",
    "H361e": "May damage the unborn child",
    "H361f": "Suspected of damaging fertility",
    "H361g": "may damage fertility",
    "H362": "May cause harm to breast-fed children",
    "H370": "Causes damage to organs",
    "H371": "May cause damage to organs",
    "H372": "Causes damage to organs through prolonged or repeated exposure",
    "H373": "May cause damage to organs through prolonged or repeated exposure",
}


def parse_ghs_classification_for_odor(
    ghs_info,
    codes=("H330", "H331", "H334", "H340", "H350", "H350i", "H351", "H36", "H37"),
    only_percent=True,
    code_only=True,
):
    strings = []
    if "Record" in ghs_info:
        for block in ghs_info["Record"]["Section"][0]["Section"][0]["Section"][0]["Information"]:
            if block["Name"] == "GHS Hazard Statements":
                for entry in block["Value"]["StringWithMarkup"]:
                    string = entry["String"]
                    for code in codes:
                        match = (code + " (") if only_percent else code
                        if match in string:
                            if code_only:
                                string = string.split(":")[0]
                            strings.append(string)
    return strings



import pandas as pd

# Load data about the odorants (names and PubChem IDs)
def get_translater():
    rosetta = pd.read_excel("SmilesInfo2.xlsx")
    threeletter_to_pubchem = dict(rosetta.iloc[:, -1:-3:-1].values)
    threeletter_to_pubchem["MEN"] = threeletter_to_pubchem["+MEN"]
    return threeletter_to_pubchem
    #pubchem_to_threeletter = {value: key for key, value in threeletter_to_pubchem.items()}
    #pubchem_to_threeletter[1201521] = "+FCH"
    #del pubchem_to_threeletter[16213045]

def get_raw():
    # Load raw mouse data (individual mouse level)
    raw = pd.read_csv(
    "raw behavioral scores mouse 73 odorants.csv", index_col=0, header=1
    ).dropna()
    raw.index.name = "odor"
    raw = raw.sort_index()
    return raw

def main():
    translater = get_translater()
    raw = get_raw()
    


import io
import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.graph_objs as go
from ipywidgets import Image, Layout, VBox

import pyrfume

try:
    from rdkit import Chem
    from rdkit.Chem import Draw
except ImportError:
    warnings.warn(
        "Parts of rdkit could not be imported; try installing rdkit via conda", UserWarning
    )


def mpl_embedding(
    xy, colors=None, alpha=0.25, figsize=(6, 6), s=0.5, cmap="hsv", title=None, ax=None
):
    if not ax:
        plt.figure(figsize=figsize)
        plt.margins(0)
        ax = plt.gca()

    ax.scatter(
        xy[:, 0],
        xy[:, 1],
        c=(colors if colors is not None else "k"),  # set colors of markers
        cmap=cmap,  # set color map of markers
        alpha=alpha,  # set alpha of markers
        marker="o",  # use smallest available marker (square)
        s=s,  # set marker size. single pixel is 0.5 on retina,
        # 1.0 otherwise
        lw=0,  # don't use edges
        edgecolor="",
    )  # don't use edges

    # remove all axes and whitespace / borders
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.set_facecolor("white")
    if title:
        ax.set_title(title)


def plotly_embedding(embedding, features=None, show_features=None, colors=None, colorscale='rainbow'):
    """
    params:
        embedding: A dataframe wrapped around e.g. a fitted TSNE object, with an index of CIDs
        features: A dataframe of features, e.g. names, SMILES strings, or physicochemical features,
            with an index of CIDs
    """

    if features is None:
        features = pyrfume.load_data("odorants/all-cids-properties.csv", usecols=range(5))
        # Only retain those rows corresponding to odorants in the embedding
    features = features.loc[embedding.index]
    show_features = show_features or list(features)
    def format_features(col):
        return "%s: %s" % (index_name, x.values.split('<br>'))
    try:
        index_name = features.index.name or 'Index'
        names = (
            features.loc[:, show_features]
            .reset_index()
            .astype("str")
            .apply(format_features, axis=1)
        )
    except Exception:
        names = features.index
    assert embedding.shape[0] == features.shape[0]

    # The scatter plot
    scatter = go.Scatter(
        x=embedding.iloc[:, 0],
        y=embedding.iloc[:, 1],
        text=names,
        mode="markers",
        hoverinfo="text",
        opacity=0.5,
        marker={
            "size": 5,
            "line": {"width": 0.5, "color": "white"},
            "color": colors if colors is not None else "black",
            "colorscale": colorscale,
        },
    )
    
    # The axes, etc.
    layout = go.Layout(
        xaxis={"type": "linear", "title": "", "showline": False, "showticklabels": False},
        yaxis={"type": "linear", "title": "", "showline": False, "showticklabels": False},
        margin={"l": 40, "b": 40, "t": 10, "r": 10},
        legend={"x": 0, "y": 1},
        hovermode="closest",
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
        width=500,
        height=500,
    )

    fig = go.FigureWidget(data=[scatter], layout=layout)
    fig.layout.hovermode = 'closest'

    # The 2D drawing of the molecule
    image_widget = Image(
        value=smiles_to_image("CCCCO"), layout=Layout(height="300px", width="300px")
    )

    def hover_fn(trace, points, state):
        ind = points.point_inds[0]
        smiles = features["SMILES"].iloc[ind]
        image_widget.value = smiles_to_image(smiles)

    scatter = fig.data[0]
    scatter.on_hover(hover_fn)
    canvas = VBox([fig, image_widget])
    return canvas



import platform
import unittest

import pyrfume


class DataLoadRemoteTestCase(unittest.TestCase):
    def test_load_manoel_2021(self):
        data = pyrfume.load_data('manoel_2021/behavior.csv')
    
    def test_load_snitz_2013(self):
        data = pyrfume.load_data('snitz_2013/behavior.csv')
        molecules = pyrfume.load_data('snitz_2013/molecules.csv')
        
    def test_load_ravia_2020(self):
        data = pyrfume.load_data('ravia_2020/behavior1.csv')
        manifest = pyrfume.load_manifest('ravia_2020')
        
    @unittest.skipIf('spike' not in platform.node(), "Only tested on Spike")
    def test_load_morgan_skipped(self):
        my_cids = [129, 239]
        morgan_sim = pyrfume.load_data('morgan/features_sim.csv', cids=my_cids)
        self.assertEqual(morgan_sim.shape[0], len(my_cids))
        
if __name__ == '__main__':
    unittest.main()



import pandas as pd
from scipy.spatial import distance_matrix
from scipy.spatial.distance import pdist, squareform
from scipy.stats import pearsonr, spearmanr


def get_distance_matrix(features):
    # Compute distance matrix.
    dist = pdist(features, metric="euclidean")
    dist = squareform(dist)
    distances = pd.DataFrame(dist, index=features.index, columns=features.index)
    return distances


def embedding_distance_correlation(original_distances, embedding, quiet=False):
    """
    params:
        original_distances: A pandas DataFrame of distances between objects
        embedding: e.g. a fitted TSNE object
    """
    transformed_distances = distance_matrix(embedding.embedding_, embedding.embedding_)
    transformed_distances = transformed_distances.ravel()
    original_distances = original_distances.values.ravel()
    r, p_r = pearsonr(original_distances, transformed_distances)
    rho, p_rho = spearmanr(original_distances, transformed_distances)
    if not quiet:
        print("R = %.3f; rho = %.3f" % (r, rho))
    return r, rho



"""Classes for odorants, mixtures, chemical orders, etc."""

import base64
import io
import json
import re
import time
import warnings
from collections import OrderedDict
from datetime import datetime
from urllib.parse import quote


import numpy as np
import pandas as pd
import pubchempy as pcp
import requests
from IPython.display import display
from PIL import Image

import quantities as pq
from pyrfume import load_data, logger, tqdm, trange
from pyrfume.physics import mackay
from quantities.constants.statisticalmechanics import R
from typing import Dict
try:
    from rdkit import Chem
    from rdkit.Chem import Draw, AllChem, SaltRemover
    from rdkit import RDLogger
    rdkit_logger = RDLogger.logger()
    RDKIT = True
except ImportError:
    warnings.warn(
        "Parts of mordred and/or rdkit could not be imported; try installing rdkit via conda",
        UserWarning,
    )
    RDKIT = False

ROOM_TEMP = (22 + 273.15) * pq.Kelvin
ROOM_PRESSURE = 1 * pq.atm
GAS_MOLAR_DENSITY = ROOM_PRESSURE / (R * ROOM_TEMP)

ODORANTS_BASIC_INFO_PATH = "molecules/all-cids-properties.csv"
ODORANT_SOURCES_PATH = "molecules/all-cids.csv"

PUBCHEM_KINDS = ['name', 'smiles', 'inchi', 'inchikey', 'formula', 'sdf', None]


class Solution:
    components: Dict["Compound", pq.quantity.Quantity] = None
    date_created: datetime = None

    def __init__(self, components: dict, date_created: datetime=None):
        self.total_volume = 0 * pq.mL
        assert isinstance(components, dict), "Components must be a dict"
        for component, volume in components.items():
            assert isinstance(
                component, (Compound, Solution)
            ), "Each component must be a Compound or a Solution"
            try:
                volume = volume.rescale(pq.mL)
            except ValueError:
                raise ValueError("Components must be provided with volumes")
            self.total_volume += volume  # Assume that volume is conserved
        self.components = components
        if not date_created:
            date_created = str(datetime.now())[:-7]
        self.date_created = date_created if date_created else datetime.now()

    @property
    def compounds(self):
        return self._compounds()

    def _compounds(self, result: dict = None):
        if result is None:
            result = {}
        for component, volume in self.components.items():
            if isinstance(component, Compound):
                if component in result:
                    result[component] += volume
                else:
                    result[component] = volume
            else:  # If it is a Solution
                component._compounds(result=result)
        return result

    @property
    def molecules(self):
        """Returns a dictionary with the moles of each Molecule"""
        compounds = self.compounds
        assert all([c.density for c, v in compounds.items() if v and not c.is_solvent]), (
            "All non-solvent compounds must have a known density " "in order to compute moles"
        )
        assert all([c.molecular_weight for c, v in compounds.items() if v and not c.is_solvent]), (
            "All non-solvent compounds must have a known molecular weight "
            "in order to compute moles"
        )
        return {
            c.molecule: (v * c.molarity).rescale(pq.mol) for c, v in self.compounds.items() if v
        }

    @property
    def molarities(self):
        """Returns a dictionary with the molarity of each Molecule"""
        return {m: mol / self.total_volume for m, mol in self.molecules.items() if mol}

    @property
    def mole_fractions(self):
        """Returns a dictionary with the mole fraction of each Molecule"""
        molecules = self.molecules
        assert [moles for molecule, moles in molecules.items()], (
            "All compounds must have a known number of moles " "in order to compute mole fraction"
        )
        # A Quantities bug prevents me from simply summing molecules.values()
        total_moles = 0 * pq.mol
        for moles in molecules.values():
            total_moles += moles
        return {molecule: moles / total_moles for molecule, moles in molecules.items()}

    def mole_fraction(self, molecule):
        return self.mole_fractions[molecule] if molecule in self.mole_fractions else 0

    @property
    def dilutions(self):
        return {
            c.molecule: self.total_volume / c.volume for c in self.compounds if not c.is_solvent
        }

    @property
    def partial_pressures(self):
        """Computes partial pressures for each odorant
        in the mixture using Raoult's law"""
        return {
            m: self.mole_fraction(m) * m.vapor_pressure for m in self.molecules if m.vapor_pressure
        }

    def partial_pressure(self, molecule):
        return self.partial_pressures[molecule]

    @property
    def total_pressure(self):
        """Computes total pressure of the vapor using Dalton's law"""
        preferred_units = pq.Pa
        partial_pressures = [pressure.rescale(preferred_units) for pressure in self.partial_pressures.values()]
        return preferred_units * np.sum(partial_pressures)

    @property
    def vapor_concentrations(self):
        """Concentrations of each component in the vapor phase at steady state.
        Units are fraction of volume. Air is assumed to make up the balance"""
        pp = self.partial_pressures
        result = {}
        for m, p in pp.items():
            ratio = (p / pq.atm).simplified
            assert ratio.units == pq.dimensionless
            result[m] = float(ratio)
        return result

    def vapor_concentration(self, molecule):
        return self.vapor_concentrations[molecule]

    @property
    def molar_evaporation_rates(self):
        mf = self.mole_fractions
        result = {
            molecule: mole_fraction * molecule.molar_evaporation_rate
            for molecule, mole_fraction in mf.items()
        }
        return result

    def molar_evaporation_rate(self, molecule):
        return self.molar_evaporation_rates[molecule]


class Molecule:
    def __init__(self, cid: int, name: str=None, fill: bool=False):
        self.cid = cid

        if fill:
            self.fill_details()
        if name:
            self.name = name

    # Integer Chemical ID number (CID) from PubChem
    cid: int = 0
    # Chemical Abstract Service (CAS) number
    cas: str = ""
    # Principal name
    name: str = ""
    # Synonyms
    synonyms: str = ""
    # IUPAC name (long, unique name)
    iupac: str = ""
    # Density (pq.g / pq.ml)
    density: float = None
    # Vapor pressure (pq.Pa)
    vapor_pressure: float = None
    # Molecular weight (pq.g / pq.mol)
    molecular_weight: float = None

    @property
    def molarity(self):
        if not (self.molecular_weight and self.density):
            result = None
        else:
            result = self.density / self.molecular_weight
            result = result.rescale(pq.mol / pq.L)
        return result

    @property
    def molar_evaporation_rate(self):
        return mackay(self.vapor_pressure)

    def fill_details(self):
        assert self.cid is not None
        url_template = (
            "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/cid/%d/property/" "%s/JSON"
        )
        property_list = ["MolecularWeight", "IsomericSMILES"]
        url = url_template % (self.cid, ",".join(property_list))
        json_data = url_to_json(url)
        details = json_data["PropertyTable"]["Properties"][0]

        def convert(name):
            s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
            return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()

        for key, value in details.items():
            if key == "CID":
                assert value == self.cid, "REST API CID does not match provided CID"
            key = convert(key)
            if key == "molecular_weight":
                value = float(value) * pq.g / pq.mol
            setattr(self, key, value)

        if not self.name:
            self.name = self.get_name_from_api()

    def get_name_from_api(self):
        url_template = "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/cid/%d/synonyms/JSON"
        url = url_template % (self.cid)
        json_data = url_to_json(url)
        name = None
        if json_data:
            information = json_data["InformationList"]["Information"][0]
            synonyms = information["Synonym"]
            name = synonyms[0].lower()
        return name

    def get_cid_from_api(self):
        url_template = "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/%s/%s/cids/JSON"
        options = [getattr(self, x) for x in ("cas", "name") if len(getattr(self, x))]
        cid = None
        query = self.name
        for option in options:
            url = url_template % (option, query)
            json_data = url_to_json(url)
        cid = json_data["IdentifierList"]["CID"][0]
        return cid

    def __eq__(self, other):
        if self.cid:
            return self.cid == other.cid
        else:
            return self.name == self.name

    def __lt__(self, other):
        if self.cid:
            return self.cid < other.cid
        else:
            return self.name < self.name

    def __hash__(self):
        return id(self)

    def __repr__(self):
        if self.cid and self.name:
            result = "%d (%s)" % (self.cid, self.name)
        elif self.cid:
            result = "%d" % self.cid
        elif self.name:
            result = "%s" % self.name
        else:
            result = "Unknown"
        return result


class Vendor:
    def __init__(self, name: str, url: str):
        self.name = name
        self.url = url

    name: str = ""
    url: str = ""


class ChemicalOrder:
    def __init__(self, molecule: "Molecule", vendor: "Vendor", part_id: str, purity: float=1, known_impurities: list=None):
        self.molecule = molecule
        self.vendor = vendor
        self.part_id = part_id
        self.purity = purity
        self.known_impurities = known_impurities

    # Molecule
    molecule: Molecule = None
    # Vendor, e.g. Sigma-Aldrich
    vendor: Vendor = None
    # ID number of compound at vendor
    part_id: str = ""
    # Reported purity as a fraction
    purity: float = 1
    # List of known impurities (Molecules)
    known_impurities: list = None


class Compound:
    def __init__(
        self, chemical_order: ChemicalOrder, stock: str="", date_arrived: datetime=None, 
        date_opened: datetime=None, is_solvent: bool=False
    ):

        self.chemical_order = chemical_order
        self.stock = stock
        self.date_arrived = date_arrived if date_arrived else datetime.now
        self.date_opened = date_opened
        self.is_solvent = is_solvent

    # ChemicalOrder
    chemical_order: ChemicalOrder = None
    # Stock number (supplied by vendor, usually on bottle)
    stock: str = ""
    # Date arrived at the lab/clinic
    date_arrived: datetime = None
    # Date opened
    date_opened: datetime = None
    # Is it a solvent?
    is_solvent: bool = False

    def __getattr__(self, attr):
        """If no attribute is found, try looking up on the
        ChemicalOrder or the Molecule"""
        try:
            return getattr(self.chemical_order, attr)
        except AttributeError:
            return getattr(self.chemical_order.molecule, attr)


def url_to_json(url, verbose=True) -> str:
    json_data = None
    response = requests.get(url)
    if response.status_code == 200:
        string = response.content.decode("utf-8")
        json_data = json.loads(string)
    else:
        msg = "HTTP Status Code %d for %s" % (response.status_code, url)
        print(msg)
        if verbose:
            logger.error(msg)
    return json_data


def is_kind(identifier: str, kind: str) -> bool:
    if kind == 'smiles':
        if RDKIT:
            rdkit_logger.setLevel(RDLogger.CRITICAL)
            result = Chem.MolFromSmiles(identifier) is not None
            rdkit_logger.setLevel(RDLogger.WARNING)
        else:
            result = None
    elif kind == 'inchikey':
        result = len(identifier) == 27 and identifier[14]=='-' and identifier[25]=='-'
    elif kind == 'inchi':
        if RDKIT:
            result = Chem.inchi.MolFromInchi(identifier, logLevel=None) is not None
        else:
            result = None
    elif kind == 'name':
        result = True
    else:
        result = False
    return result

    
def get_kind(identifier: str):
    kinds = [kind for kind in PUBCHEM_KINDS if is_kind(identifier, kind)]
    return kinds[-1]  # Return most sophisticated kind ('name' will always be in the list)
        

def deisomerize_smiles(smiles: str) -> str:
    mol = Chem.MolFromSmiles(smiles)
    if mol:  # If a mol object was successfully create (i.e. not `None`)
        smiles = Chem.MolToSmiles(mol, isomericSmiles=False)
    else:
        smiles = smiles.replace('@','').replace('@@','').replace('/','').replace('\\','')
    return smiles


def canonical_smiles(smiles: str, kekulize: bool = False) -> str:
    """Use rdkit to convert the `smiles` string to canonical form"""
    mol = Chem.MolFromSmiles(smiles)
    if mol:  # If a mol object was successfully create (i.e. not `None`)
        if kekulize:
            Chem.Kekulize(mol)
        smiles = Chem.MolToSmiles(mol, isomericSmiles=True)
    else:  # No mol object means the `smiles` string was invalid
        smiles = ""
    return smiles


def get_cids(
    identifiers: list,
    kind: str = None,
    verbose: bool = True,
    wait: float = 0,
    results: dict = None,
) -> dict:
    """Return CIDs for molecule based on any synonym,
    including a chemical name or a CAS"""
    if isinstance(kind, str):
        kind = kind.lower()
    assert kind in PUBCHEM_KINDS
    if results is None:
        results = {}
    p = tqdm(identifiers)
    for identifier in p:
        #if not isinstance(identifier, str):
        #    logger.warning("%s is not a string" % identifier)
        #    continue
        p.set_description(str(identifier))
        cid = get_cid(identifier, kind=kind, verbose=verbose)
        if not cid:
            logger.warning("Could not find %s" % identifier)
        results[identifier] = cid
        if wait:
            time.sleep(wait)
    return results


def get_cid(
    identifier: str, kind: str = None, verbose: bool = True, fix_smiles_on_error: bool = True, attempt=0
) -> int:
    """
    Return data about a molecule from any synonym,
    including a chemical name or a CAS.
    """
    if isinstance(identifier, float) and np.isnan(identifier):
        return 0
    replace = [('α', 'alpha'), ('β', 'beta'), ('γ', 'gamma'), ('δ', 'delta')]
    for a, b in replace:
        identifier = identifier.replace(a, b)
    if kind is None:
        kind = get_kind(identifier)
    else:
        kind = kind.lower()
    try:
        result = pcp.get_cids(identifier, namespace=kind)
    except pcp.BadRequestError:
        logger.warning('Request Error for "%s"' % identifier)
        result = []
    except pcp.PubChemHTTPError as e:
        if attempt == 0:
            import time
            time.sleep(10)
            return get_cid(identifier,kind,verbose, fix_smiles_on_error, 1)
        else:
            raise e
    if not len(result):
        cid = 0
    else:
        if (len(result) > 1) and verbose:
            logger.warning("Multiple CIDs for %s: %s" % (identifier, result))
        cid = result[0]
    if not cid and kind == "smiles" and fix_smiles_on_error:
        # Retry with canonical SMILES
        identifier = canonical_smiles(identifier)
        if identifier:
            cid = get_cid(identifier, kind=kind, verbose=verbose, fix_smiles_on_error=False)
    return cid


def from_cids(cids: list, property_list: bool = None) -> list:
    if property_list is None:
        property_list = ["MolecularWeight", "IsomericSMILES", "IUPACName"]
    result = []
    chunk_size = 100
    for start in trange(0, len(cids), chunk_size):
        stop = min(start + chunk_size, len(cids))
        logger.info("Retrieving %d through %d" % (start, stop - 1))
        cid_subset = [
            str(x) for x in [cid for cid in cids[start:stop] if int(cid) > 0 and cid is not None]
        ]
        cid_subset = ",".join(cid_subset)
        properties_template = (
            "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/cid/%s/property/" "%s/JSON"
        )
        url = properties_template % (cid_subset, ",".join(property_list))
        json_data = url_to_json(url)
        data = json_data["PropertyTable"]["Properties"]

        synonyms_template = (
            "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/cid/%s/synonyms/JSON"
        )
        url = synonyms_template % (cid_subset)
        json_data = url_to_json(url)
        information = json_data["InformationList"]["Information"]
        for i, d in enumerate(data):
            try:
                synonyms = information[i]["Synonym"]
            except KeyError:
                try:
                    d["name"] = d["IUPACName"]
                except KeyError:
                    d["name"] = ""
            else:
                d["name"] = synonyms[0].lower()
        result += data
    return result


def cids_to_smiles(cids: list) -> dict:
    """Returns an ordered dictionary of SMILES strings with CIDs as keys"""
    info = from_cids(cids, property_list=["IsomericSMILES"])
    smiles = {item["CID"]: item["IsomericSMILES"] for item in info}
    return smiles


def cids_to_cas(cids: list) -> OrderedDict:
    result = OrderedDict()
    chunk_size = 100
    for start in trange(0, len(cids), chunk_size):
        stop = min(start + chunk_size, len(cids))
        # msg = "Retrieving %d through %d" % (start, stop-1)
        cid_subset = [str(x) for x in cids[start:stop]]
        cid_subset = ",".join(cid_subset)
        synonyms_template = (
            "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/cid/%s/synonyms/JSON"
        )
        url = synonyms_template % (cid_subset)
        json_data = url_to_json(url)
        information = json_data["InformationList"]["Information"]
        for i, info in enumerate(information):
            cid = info["CID"]
            try:
                synonyms = info["Synonym"]
            except KeyError:
                result[cid] = []
            else:
                result[cid] = cas_from_synonyms(synonyms)
        result
    return result


def cas_from_synonyms(synonyms: list) -> list:
    result = []
    for s in synonyms:
        if re.match(r"^[0-9]+\-[0-9]+\-[0-9]+$", s):
            result.append(s)
    return result


def cactus(identifier: str, output: str = "cas") -> str:
    url_template = "https://cactus.nci.nih.gov/chemical/structure/%s/%s"
    identifier = identifier.replace(' ', '%20')
    url = url_template % (identifier, output)
    response = requests.get(url)
    if response.status_code == 200:
        result = response.content.decode("utf-8")
    else:
        logger.error("HTTP Status Code %d for %s" % (response.status_code, url))
        result = None
    return result


def cactus_image(smiles: str) -> None:
    url_template = "https://cactus.nci.nih.gov/chemical/structure/%s/image"
    smiles = smiles.replace(' ', '%20')
    url = url_template % smiles
    response = requests.get(url)
    if response.status_code == 200:
        image_data = response.content.decode("utf-8")
        image = Image(image_data)
        display(image)
    else:
        logger.error("HTTP Status Code %d for %s" % (response.status_code, url))


def get_compound_summary(cid: int, heading: str):
    """Get summary info about `heading` from PubChem for the compound
    given by `cid`.  Example heading: 'Physical Description'"""
    url_template = (
        "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug_view/data/compound/%d/JSON?heading=%s"
    )
    escaped_heading = quote(heading)  # Escape the string
    url = url_template % (cid, escaped_heading)
    json_data = url_to_json(url, verbose=False)
    return json_data


def get_compound_odor(cid, raw=False):
    info = []
    for heading in ["Odor", "Physical Description"]:
        json_data = get_compound_summary(cid, heading)
        if raw:
            info += [] if json_data is None else [json_data]
        else:
            info += _parse_odor_info(json_data)
    return info


def _parse_odor_info(info, odors=None, any_string=False):
    if odors is None:
        odors = []
    if isinstance(info, dict):
        for key, value in info.items():
            if key == "TOCHeading" and value == "Odor":
                any_string = True
            if key == "String" and (any_string or "odor" in value.lower()):
                odors.append(value)
            else:
                _parse_odor_info(value, odors=odors, any_string=any_string)
    elif isinstance(info, list):
        for value in info:
            _parse_odor_info(value, odors=odors, any_string=any_string)
    return odors


def _parse_other_info(info, records=None):
    if records is None:
        records = []
    if isinstance(info, dict):
        for key, value in info.items():
            if key == "String":
                records.append(value)
            elif key == "Value" and "Number" in value:
                records.append(value)
            else:
                _parse_other_info(value, records=records)
    elif isinstance(info, list):
        for value in info:
            _parse_other_info(value, records=records)
    return records


def display_molecules(molecules: pd.DataFrame, no_of_columns=5, figsize=(15, 15)):
    import matplotlib.pyplot as plt
    from IPython.display import display
    fig = plt.figure(figsize=figsize)
    column = 0
    for i, (cid, info) in enumerate(molecules.iterrows()):
        column += 1
        #  check for end of column and create a new figure
        if column == no_of_columns+1:
            fig = plt.figure(figsize=figsize)
            column = 1
        fig.add_subplot(1, no_of_columns, column)
        image = smiles_to_image(info['IsomericSMILES'], png=False)
        plt.imshow(image)
        plt.axis('off')
        plt.title("%d: %s" % (cid, info['name']))
        
        
def embed_molecules(molecules: pd.DataFrame):
    import matplotlib.pyplot as plt
    plt.figure(figsize=(6, 6))
    ax = plt.gca()
    embedding = load_data('embedding/pf_umap.pkl')
    ax = embedding.plot.scatter(x=0, y=1, alpha=0.05, c='k', ax=ax)
    smiles = molecules['IsomericSMILES']
    embedding_ = embedding.loc[smiles]
    embedding_.plot.scatter(x=0, y=1, alpha=1, c='r', s=100, ax=ax)
    ax.set_xlabel('Dimension 1')
    ax.set_ylabel('Dimension 2')
    


def smiles_to_image(smiles, png=True, b64=False, crop=True, padding=10, size=300):
    """
    png: Whether to convert to .png data (or to leave as a PIL image)
    b64: Whether to base64 encode (only possible for .png data)
    """
    buffer = io.BytesIO()
    mol = Chem.MolFromSmiles(smiles)
    image = Draw.MolToImage(mol, fitImage=True, size=(size, size))
    if crop:
        image = crop_image(image, padding=padding)
    if png:
        image.save(buffer, format="PNG")
        image = buffer.getvalue()
    if b64:
        assert png, "Can only base64 encode PNG data, not a PIL image"
        image = base64.b64encode(image).decode("utf8")
    return image


def smiles_to_mol(smiles: list, max_attempts: int=10, use_random_coords: bool=False, deisomerize=False) -> dict:
    if deisomerize:
        f = deisomerize_smiles
    else:
        f = lambda x: x
    mols_raw = [Chem.MolFromSmiles(f(smi)) for smi in smiles]
    logger.info("Computing 3D coordinates...")
    s = SaltRemover.SaltRemover()
    mols = {}
    n = len(mols_raw)
    pbar = tqdm(total=n)
    for i, mol in enumerate(mols_raw):
        pbar.update()
        logger.debug("Embedding %s" % smiles[i])
        try:
            mol = s.StripMol(mol, dontRemoveEverything=True)
            mol = Chem.AddHs(mol)
            AllChem.Compute2DCoords(mol)
            AllChem.EmbedMolecule(mol, maxAttempts=max_attempts, useRandomCoords=use_random_coords)
            AllChem.UFFOptimizeMolecule(mol)  # Is this deterministic?
        except Exception as e:
            logger.warning("Exception for %s: %s" % (smiles[i], str(e)))
        else:
            mols[smiles[i]] = mol
    logger.info("Finished embedding all molecules")
    return mols


def crop_image(img, padding=0):
    """Crop white out of a PIL image."""
    as_array = np.array(img)  # N x N x (r,g,b,a)
    if as_array.shape[2] == 4:
        as_array[as_array[:, :, 3] == 0] = [255, 255, 255, 255]
    has_content = np.sum(as_array, axis=2, dtype=np.uint32) != 255 * 4
    xs, ys = np.nonzero(has_content)
    x_range = max([min(xs) - padding, 0]), min([max(xs) + padding, as_array.shape[0]])
    y_range = max([min(ys) - padding, 0]), min([max(ys) + padding, as_array.shape[1]])
    as_array_cropped = as_array[x_range[0] : x_range[1], y_range[0] : y_range[1], 0:3]
    img = Image.fromarray(as_array_cropped, mode="RGB")
    return img


def all_odorants():
    """All CIDs, SMILES, Names, and Molecular Weights found in the
    file at ODORANTS_BASIC_INFO_PATH"""
    df = load_data(ODORANTS_BASIC_INFO_PATH)
    df = df.sort_index()
    return df


def all_sources():
    """Whether or not each odorant (by CID) is in each of the data sources"""
    df = load_data(ODORANT_SOURCES_PATH)
    df = df.sort_index()
    return df


def all_cids():
    """All CIDs found in the file at ODORANTS_BASIC_INFO_PATH"""
    df = all_odorants()
    return list(df.index)


def all_smiles():
    """All SMILES found in the file at ODORANTS_BASIC_INFO_PATH.
    May contain duplicates (if two CIDs give the same SMILES)"""
    df = all_odorants()
    return list(df["SMILES"])


if __name__ == "__main__":
    x = Molecule(325, fill=True)
    print(x.__dict__)



"""Module for interacting with Haddad et al, 2008"""

import pandas as pd
from scipy.spatial.distance import pdist, squareform
from sklearn.preprocessing import StandardScaler

import pyrfume
from . import features
from .base import DEFAULT_DATA_PATH

HADDAD_DIR = DEFAULT_DATA_PATH / "haddad_2008"


def get_haddad_dragon(use_original=True, regenerate=False):
    path = HADDAD_DIR / "haddad_dragon.csv"
    if not path.is_file() or regenerate:
        haddad_feature_names = get_haddad_weights().index
        # Use standard scaling as in the Snitz paper
        standard_scaled_dragon = features.load_dragon(suffix="-cleaned-standardized-imputed")
        df = standard_scaled_dragon[haddad_feature_names]
        pyrfume.save_data(df, path)
    else:
        df = pyrfume.load_data(path)  # .set_index('PubChemID')
    return df


def get_haddad_weights():
    """Get a pandas Series of weights for Dragon features in Haddad"""
    # haddad_info_path = HADDAD_DIR / 'haddad-optimized_v6-symbols.xlsx'
    # df_haddad_list = pd.read_excel(haddad_info_path)
    # haddad_list = df_haddad_list['Dragon 6.0 symbol']
    # haddad_weights = df_haddad_list['Weight']
    # haddad_weights.index = haddad_list
    path = HADDAD_DIR / "haddad-weights-from-paper-dragon-6.csv"
    haddad_weights = pyrfume.load_data(path, index_col=0)["Weight"]
    return haddad_weights


def get_haddad_features(dragon, weights=None):
    """Extract weighted features from the dataframe `dragon`"""
    if not weights:
        haddad_weights = get_haddad_weights()

    # Scale dragon dataframe (may already be scaled)
    ss = StandardScaler()
    scaled = ss.fit_transform(dragon[haddad_weights.index])
    haddad_features = pd.DataFrame(scaled, index=dragon.index, columns=haddad_weights.index)

    # Multiply by weights
    haddad_features = haddad_features.mul(haddad_weights, axis=1)

    return haddad_features


def get_haddad_distances(dragon, weights=None):
    haddad_features = get_haddad_features(dragon, weights=weights)
    # Make Haddad distance matrix
    x = pdist(haddad_features.values, "euclidean")
    haddad_distances = pd.DataFrame(
        squareform(x), index=haddad_features.index, columns=haddad_features.index
    )

    return haddad_distances



import sys
import unittest

from .test_docs import *
from .test_load_data import *
from .test_config_data import *
from .test_make_solutions import *
from .test_mixture import *
from .test_test import *
from .test_others_in__init__ import *
from .test_physics import *
from .test_odorants import *
from .test_datajoint_tools import DataJointTestCase

def main():
    buffer = 'buffer' in sys.argv
    database = 'database' in sys.argv
    sys.argv = sys.argv[:1]

    if not database:
        globals().pop('DataJointTestCase')

    unittest.main(buffer=buffer)

if __name__ == '__main__':
    main()



import os

import pandas as pd
from sklearn.preprocessing import StandardScaler

from missingpy import KNNImputer

from .base import DATA_DIR


def clean_dragon(save=False):
    source = os.path.join(DATA_DIR, "cids-smiles-dragon.txt")
    df = pd.read_csv(source).set_index("CID")
    df = df.iloc[:, 1:]  # Drop SMILES column

    # Scale to mean 0, variance 1
    ss = StandardScaler()
    good = df.columns[df.isnull().sum() < 500]
    df = df[good]
    scaled = ss.fit_transform(df.astype("float"))
    df = pd.DataFrame(scaled, index=df.index, columns=df.columns)

    # Impute missing values
    knn = KNNImputer(k=5)
    imputed = knn.fit_transform(df.values)
    df = pd.DataFrame(imputed, index=df.index, columns=df.columns)

    # Optionally save to disk
    if save:
        dest = os.path.join(DATA_DIR, "cids-smiles-dragon-scaled-imputed.txt")
        df.to_csv(dest)

    return df



import unittest
import datajoint as dj
import pandas as pd



class DataJointTestCase(unittest.TestCase):

    def setUp(self):
        dj.config['database.host'] = '127.0.0.1'
        dj.config['database.user'] = 'root'
        dj.config['database.password'] = 'simple'

    def test_schematizing_odorants(self):
        """Test case for schematizations of classes in the odorants module 
            and inserting sample data into datajoint table.
        """

        from pyrfume.odorants import Molecule, Vendor, ChemicalOrder, Compound, Solution
        from pyrfume import datajoint_tools as djt

        schematized_classes = {}
        schema = dj.schema("test_schema")
        schema.drop(force=True)
        schema = dj.schema("test_schema", schematized_classes)

        classes_tobe_schematized = ['Molecule', 'Vendor', 'ChemicalOrder', 'Compound', 'Solution']
        for cls_name in classes_tobe_schematized:
            schematized_classes[cls_name] = djt.schematize(locals()[cls_name], schema)

        # test if definition defined.
        for cls_name in classes_tobe_schematized:
            self.assertIsInstance(schematized_classes[cls_name].definition, str)


        Molecule = schematized_classes['Molecule']
        Molecule.insert1({'cid': 6334, 'cas': '', 'name': 'propane', 'iupac': 'propane', 'synonyms': ''})
        molecules = pd.DataFrame(Molecule.fetch())
        self.assertEqual(molecules.shape[0], 1)

        Vendor = schematized_classes['Vendor']
        Vendor.insert1({"name": "vendor1", "url": "www.vendor1.com"})
        vendors = pd.DataFrame(Vendor.fetch())
        self.assertEqual(vendors.shape[0], 1)

        ChemicalOrder = schematized_classes['ChemicalOrder']
        ChemicalOrder.insert1({"molecule_id": 1, "vendor_id": 1, 'part_id': 1, 'purity': 0.5})
        chemicalorders = pd.DataFrame(ChemicalOrder.fetch())
        self.assertEqual(chemicalorders.shape[0], 1)

        Compound = schematized_classes['Compound']
        Compound.insert1({'chemicalorder_id': 1, 'stock': 'enough', 'is_solvent': 1, 'date_arrived': '2020-09-27 02:18:24', 'date_opened': '2020-09-27 02:18:24'})
        compounds = pd.DataFrame(Compound.fetch())
        self.assertEqual(compounds.shape[0], 1)

        Solution = schematized_classes['Solution']
        Solution.insert1({"date_created": '2020-09-27 02:18:24'})
        solutions = pd.DataFrame(Solution.fetch())
        self.assertEqual(solutions.shape[0], 1)

        Solution.Component.insert1({"solution_id": 1, "compound_id": 1, "value": 1.0})
        solution_components = pd.DataFrame(Solution.Component.fetch())
        self.assertEqual(solution_components.shape[0], 1)


if __name__ == '__main__':
    unittest.main()



import numpy as np

import pyrfume


def get_mainland(raw=False, vendors=None):
    """Return a dataframe containing odorants in Joel Mainland's cabinet"""
    df = pyrfume.load_data("cabinets/mainland.csv").set_index("CID")
    if not raw:
        if vendors:  # e.g. ['sigma']
            # Keep only odorants obtained from certain vendors
            df = df[df["SourcedFrom"].str.lower().isin(vendors)]

        # For odorants with no relative cost information
        # (probably discontinued), fill with a large number so they
        # don't get used.  Also use log10 $/mol
        df.loc[:, "$/mol"] = np.log10(df.loc[:, "$/mol"].fillna(1e15))

        # Fix price information
        df["Price"] = df["Price"].astype("str").apply(lambda x: x.replace("$", "")).astype("float")

        # Remove odorants with no price
        df = df.dropna(subset=["Price"])

        # Sort by price and then take only the cheapest instance of each CID
        df = df.sort_values("Price").groupby("CID").first()

        # Drop values with no CID
        df = df.loc[df.index > 0]

        # Sort CIDs from low to high
        df = df.sort_index()

    return df



"""Module for interacting with unpublished data from Westeros"""

import pandas as pd
from scipy.spatial.distance import pdist, squareform

import pyrfume


def get_x_embeddings():
    emb_raw = pyrfume.load_data("westeros/embeddings.csv")
    emb = emb_raw[[x for x in list(emb_raw) if "emb_" in x]]
    emb = emb.dropna()
    return emb


def get_x_distances(embeddings):
    emb = get_x_embeddings()
    # Compute distance matrix.
    # This will give basically the same low-D embedding as using
    # raw high-dimensional embeddings does
    dist = pdist(emb, metric="euclidean")
    dist = squareform(dist)
    distances = pd.DataFrame(dist, index=emb.index, columns=emb.index)
    return distances



import configparser
import json
import logging
import numpy as np
import pandas as pd
from pathlib import Path
import pickle
from pprint import pprint
import requests
import tempfile
import toml
from tqdm.auto import tqdm, trange
from typing import Any
import urllib


PACKAGE_DIR = Path(__file__).resolve().parent
CONFIG_PATH = PACKAGE_DIR / "config.ini"
DEFAULT_DATA_PATH = PACKAGE_DIR.parent.parent / "pyrfume-data"
REMOTE_DATA_ORG = 'pyrfume'
REMOTE_DATA_REPO = 'pyrfume-data'
REMOTE_DATA_PATH = 'https://raw.githubusercontent.com/%s/%s' % (REMOTE_DATA_ORG, REMOTE_DATA_REPO)
TEMP_LOCAL = tempfile.TemporaryDirectory()
MANIFEST_NAME = 'manifest.toml'


class LocalDataError(Exception):
    pass


class RemoteDataError(Exception):
    pass


logger = logging.getLogger("pyrfume")


def init_config(overwrite=False):
    if overwrite or not CONFIG_PATH.exists():
        config = configparser.ConfigParser()
        config["PATHS"] = {"pyrfume-data": str(DEFAULT_DATA_PATH)}
        config["DATABASE"] = {"schema_name": "UNDEFINED"}
        with open(CONFIG_PATH, "w") as f:
            config.write(f)


def reset_config():
    init_config(overwrite=True)


def read_config(header, key):
    config = configparser.ConfigParser()
    init_config()
    config.read(CONFIG_PATH)
    return config[header][key]


def write_config(header, key, value):
    config = configparser.ConfigParser()
    init_config()
    config.read(CONFIG_PATH)
    config[header][key] = value
    with open(CONFIG_PATH, "w") as f:
        config.write(f)


def set_data_path(path, create=True):
    path = Path(path).resolve()
    if create:
        path.mkdir(exist_ok=True, parents=True)
    if not path.exists():
        raise Exception("Could not find path %s" % path)
    write_config("PATHS", "pyrfume-data", str(path))


def get_data_path(create=True):
    path = read_config("PATHS", "pyrfume-data")
    path = Path(path).resolve()
    if create:
        path.mkdir(exist_ok=True, parents=True)
    if not path.exists():
        raise Exception("Could not find data path %s" % path)
    return path


def get_remote_data_path(branch='master'):
    path = REMOTE_DATA_PATH + '/' + branch
    return path


def localize_remote_data(rel_path, branch='master', quiet=False):
    url = get_remote_data_path(branch=branch) + '/' + rel_path
    target_path = Path(TEMP_LOCAL.name) / rel_path
    response = requests.get(url)
    if response.status_code != 200:
        if not quiet:
            raise RemoteDataError('Could not get file at %s' % url)
        else:
            return None
    target_path.parent.mkdir(exist_ok=True)
    with open(target_path, 'wb') as f:
        f.write(response.content)
    return target_path
    
def get_remote_archives_info(branch='master'):
    url = 'https://api.github.com/repos/%s/%s/git/trees/%s' % (REMOTE_DATA_ORG, REMOTE_DATA_REPO, branch)
    response = requests.get(url)
    if response.status_code != 200:
        raise RemoteDataError('Could not get archive list at %s' % url)
    info = json.loads(response.content)
    return info

    
def list_archives(branch='master', remote=None):
    archives = []
    if remote:
        info = get_remote_archives_info(branch=branch)
        for item in info['tree']:
            if item['type'] == 'tree':
                archives.append(item['path'])    
    else:
        path = get_data_path()
        for directory in path.iterdir():
            if (directory / 'manifest.toml').is_file():
                archives.append(directory.name)
        if not len(archives):
            f = logger.info if remote is None else logger.warning
            f('No local archives found; searching remotely...')
            return list_archives(branch=branch, remote=True)
    archives = sorted(archives)
    return archives
    
    
def load_manifest(archive_name, remote=None):
    rel_path = archive_name + '/' + MANIFEST_NAME 
    return load_data(rel_path, remote=remote)


def show_files(archive_name, remote=None, raw=False):
    manifest = load_manifest(archive_name, remote=remote)
    items = manifest['processed']
    if raw:
        items.update(manifest['raw'])
    pprint(items)
    
    
def load_data(rel_path, remote=None, cids=None, quiet=False, **kwargs):
    if remote:
        full_path = localize_remote_data(rel_path, quiet=quiet)
    else:
        full_path = get_data_path() / rel_path
        if not full_path.exists():
            if remote is None:
                logger.info('Did not find file %s locally; fetching remotely.' % full_path)
                return load_data(rel_path, remote=True, quiet=quiet, **kwargs)
            elif not quiet:
                raise LocalDataError('Could not get file at %s' % full_path)
    if not full_path:
        return None
    
    is_csv = full_path.suffix in [".csv", ".txt"]
    is_pickle = full_path.suffix in [".pkl", ".pickle", ".p"]
    is_excel = full_path.suffix in  [".xls", ".xlsx"]
    is_manifest = full_path.suffix == Path(MANIFEST_NAME).suffix
    
    if is_pickle:
        with open(full_path, "rb") as f:
            data = pickle.load(f)
    elif is_excel:
        data = pd.read_excel(full_path, **kwargs)
    elif is_csv:
        if "index_col" not in kwargs:
            kwargs["index_col"] = 0
        if cids:
            # First check for a CIDs file specific to this dataset
            cids_path = str(Path(rel_path).parent / 'cids.csv')
            cids_data = load_data(cids_path, remote=remote, quiet=True)
            if cids_data is not None:
                all_cids_ = cids_data.index.tolist()
            else:
                # If not found, fall back on the generic Pyrfume list of CIDs
                from pyrfume.odorants import all_cids
                all_cids_ = all_cids() # Thousands of CIDs to choose from.
            skip_f = lambda line_num: False if not line_num else all_cids_[line_num-1] not in cids
            kwargs['skiprows'] = skip_f
        data = pd.read_csv(full_path, **kwargs)
        if cids and sorted(data.index.tolist()) != sorted(cids):
            logger.warning('CIDs returned did not match CIDs requested; does this dataset have a custom CID list?')
    elif is_manifest:
        with open(full_path, 'r') as f:
            data = toml.load(f)
    else:
        raise LocalDataError("%s has an unknown data type" % rel_path)
    return data


def save_data(data, rel_path, create_archive=True, **kwargs):
    full_path = get_data_path() / rel_path
    is_pickle = any(str(full_path).endswith(x) for x in (".pkl", ".pickle", ".p"))
    is_csv = any(str(full_path).endswith(x) for x in (".csv"))
    archive_path = full_path.parent
    if not archive_path.is_dir():
        if create_archive:
            archive_path.mkdir(exist_ok=True)
        else:
            raise LocalDataError("Archive %s does not exist; use `create_archive=True` to create it" % archive_path.name)
    if is_pickle:
        with open(full_path, "wb") as f:
            pickle.dump(data, f)
    elif is_csv:
        data.to_csv(full_path, **kwargs)
    else:
        raise Exception("Unsupported extension in file name %s" % full_path.name)



from pyrfume.odorants import Solution, Compound, ChemicalOrder, \
                                  Vendor, Molecule
import quantities as pq
import unittest

class MakeSolutionTestCase(unittest.TestCase):
    def test_make_solution(self):

        # Instantiate two molecules by CID number
        beta_pinene = Molecule(14896, fill=True)  # Fruity smell
        d_limonene = Molecule(440917, fill=True)  # Lemon smell (check)
        l_limonene = Molecule(439250, fill=True)  # Orange smell (check)
        light_mineral_oil = Molecule(347911206, fill=True)  # An odorless solvent

        # Vapor pressures at 25 degrees Celsius (obtained from PubChem)
        beta_pinene.vapor_pressure = 10 * pq.mmHg # Made up, fill with real values
        d_limonene.vapor_pressure = 10 * pq.mmHg # Made up, fill with real values
        l_limonene.vapor_pressure = 10 * pq.mmHg # Made up, fill with real values
        light_mineral_oil.vapor_pressure = 0 * pq.mmHg  # Actually .0001 * pq.hPa

        # Densities at 20 degrees Celsius (obtained from PubChem)
        beta_pinene.density = 1.0 * pq.g/pq.cc # Made up, fill with real values
        d_limonene.density = 1.0 * pq.g/pq.cc # Made up, fill with real values
        l_limonene.density = 1.0 * pq.g/pq.cc # Made up, fill with real values
        light_mineral_oil.density = 0.85 * pq.g/pq.cc # Made up, fill with real values

        light_mineral_oil.molecular_weight = 500 * pq.g / pq.mol

        sigma_aldrich = Vendor('Sigma Aldrich', 'http://www.sigma.com')

        solvent_order = ChemicalOrder(light_mineral_oil, sigma_aldrich, '')

        chemical_orders = [ChemicalOrder(beta_pinene, sigma_aldrich, ''),
                        ChemicalOrder(d_limonene, sigma_aldrich, ''),
                        ChemicalOrder(l_limonene, sigma_aldrich, '')]

        solvent = Compound(solvent_order, is_solvent=True)
        compounds = [Compound(chemical_order) for chemical_order in chemical_orders]
        n_odorants = len(compounds)

        solutions = []
        solutions.append(Solution({compounds[0]: 1 * pq.mL,
                                solvent: 24 * pq.mL}))
        solutions.append(Solution({compounds[1]: 1 * pq.mL,
                                solvent: 24 * pq.mL}))
        solutions.append(Solution({compounds[2]: 1 * pq.mL,
                                solvent: 24 * pq.mL}))
        solutions.append(Solution({compounds[0]: 0.01 * pq.mL,
                                solvent: 24 * pq.mL}))
        solutions.append(Solution({compounds[1]: 0.01 * pq.mL,
                                solvent: 24.9 * pq.mL}))
        solutions.append(Solution({compounds[2]: 0.01 * pq.mL,
                                solvent: 24.9 * pq.mL}))
        n_solutions = len(solutions)



import unittest
import numpy as np
from pyrfume.experiments import Distance, Result, odorant_distances, ROC, correct_matrix, TriangleTest
from .unittest_utils import get_substances



class OthersTestCase(unittest.TestCase):
    def setUp(self):
        self.mixtures = get_substances("mixtures")
        self.mixture_C4H8O2, self.mixture_C4H8S, self.mixture_C2H6O = self.mixtures
        self.components = get_substances("components")
        self.component_C4H8O2, self.component_C4H8S, self.component_C2H6O = self.components

    def test_Distance(self):
        distance = Distance(self.mixture_C4H8O2, self.mixture_C2H6O, 1.0)

    def test_odorant_distances_ROC_correct_matrix(self):
        test1 = TriangleTest(0, list(self.mixtures), 0.5, True)
        new_odorants = [self.mixture_C4H8O2, self.mixture_C2H6O, self.mixture_C2H6O]
        test1.add_odorants(new_odorants)
        
        test2 = TriangleTest(1, list(self.mixtures), 0.5, False)
        new_odorants = [self.mixture_C4H8O2, self.mixture_C4H8S, self.mixture_C4H8S]
        test2.add_odorants(new_odorants)

        result1 = Result(test1, 0, True)
        result2 = Result(test2, 1, False)
        results = [result1, result2]

        distances: dict = odorant_distances(results, 0)
        self.assertTrue(1 in distances.values())
        self.assertTrue(0 in distances.values())
        
        roc = ROC(results, 1)
        self.assertEqual(roc[0][0], 1)
        self.assertEqual(roc[1][0], 1)
        matrix = correct_matrix(results, 1, None)

if __name__ == '__main__':
    unittest.main()



import unittest
from pyrfume.physics import *
import quantities as pq

class PhysicsTestCase(unittest.TestCase):
    def test_mackay(self):
        q = 1 * pq.Pa
        self.assertAlmostEqual(2.8238e-07, mackay(q).item(), 3)
        q = -1 * pq.Pa
        self.assertEqual(0, mackay(q).item())
    
    def test_bernoulli(self):
        bernoulli()
        bernoulli(1 * pq.m/pq.s)
        bernoulli(1 * pq.m/pq.s, 1 * pq.Pa)
        bernoulli(1 * pq.m/pq.s, 1 * pq.Pa, 1 * pq.g / pq.cm ** 3)
        bernoulli(1 * pq.m/pq.s, 1 * pq.Pa, 1 * pq.g / pq.cm ** 3, 3.72076 * pq.m / (pq.s) ** 2)
        bernoulli(1 * pq.m/pq.s, 1 * pq.Pa, 1 * pq.g / pq.cm ** 3, 3.72076 * pq.m / (pq.s) ** 2, 1)
        self.assertEqual(
            bernoulli(
                1 * pq.m/pq.s, 
                1 * pq.Pa, 
                1 * pq.g / pq.cm ** 3, 
                3.72076 * pq.m / (pq.s) ** 2, 
                1, 
                1 * (pq.m / pq.s) ** 2
            ),
            []
        )

    def test_venturi(self):
        venturi(1 * pq.g / pq.cm ** 3)
        venturi(1 * pq.g / pq.cm ** 3, 1 * pq.Pa)
        venturi(1 * pq.g / pq.cm ** 3, 1 * pq.Pa, 2 * pq.Pa, )

        v2 = Symbol("v2", real=True, positive=True)
        result = venturi(1 * pq.g / pq.cm ** 3, 1 * pq.Pa, 2 * pq.Pa, 1 * pq.m / pq.s)[0][v2]
        self.assertAlmostEqual(result, 0.998999, 4)

        result = venturi(1 * pq.g / pq.cm ** 3, 1 * pq.Pa, 2 * pq.Pa, 1 * pq.m / pq.s, 2 * pq.m / pq.s)
        self.assertEqual(result, [])



if __name__ == '__main__':
    unittest.main()



import pandas as pd
from scipy.spatial.distance import pdist, squareform

import pyrfume

from . import features
from .base import DEFAULT_DATA_PATH

SNITZ_DIR = DEFAULT_DATA_PATH / "snitz_2013"


def get_snitz_dragon(use_original=True, regenerate=False):
    path = SNITZ_DIR / "snitz_dragon.csv"
    if not path.is_file() or regenerate:
        snitz_feature_names = get_snitz_weights().index
        # Use minmax scaling as in the Snitz paper
        minmax_scaled_dragon = features.load_dragon(suffix="-cleaned-minmaxed-imputed")
        df = minmax_scaled_dragon[snitz_feature_names]
        pyrfume.save_data(df, path)
    else:
        df = pyrfume.load_data(path)  # .set_index('PubChemID')
    return df


def get_snitz_weights(use_original=True):
    """Return a pandas Series of weights for Dragon features in Snitz"""
    if use_original:  # Use the ones from the Snitz paper, with no weights
        file_name = "snitz-descriptors-from-paper-dragon-6.csv"
        path = SNITZ_DIR / file_name
        snitz_weights = pyrfume.load_data(path, index_col=0)["Weight"]
    else:
        # Use the ones that I derived, with weights computed by optimization
        # using Snitz-space projections of each molecule's original unit vector
        file_name = "snitz_dragon_weights.csv"
        path = SNITZ_DIR / file_name
        snitz_weights = -pyrfume.load_data(path, index_col=0)["Weight"]
    return snitz_weights


def get_snitz_features(dragon, snitz_weights=None, use_original=True):
    if snitz_weights is None:
        snitz_weights = get_snitz_weights(use_original=use_original)
    snitz_features = dragon[snitz_weights.index] * snitz_weights
    # CIDs where the Snitz vector has no length
    null_vectors = list(snitz_features.index[snitz_features.sum(axis=1) == 0])
    print("Number of zero-length Snitz vectors is %d " % len(null_vectors))
    # Fill these with median features (these molecules will become useless)
    for nv in null_vectors:
        snitz_features.loc[nv, :] = snitz_features.median()
    return snitz_features


def get_snitz_distances(dragon, snitz_features=None, snitz_weights=None, use_original=True):
    # Compute Snitz distances.
    # Cosine distance is approximately the same as angle distance
    # for small-ish angles
    if snitz_features is None:
        snitz_features = get_snitz_features(
            dragon, snitz_weights=snitz_weights, use_original=use_original
        )
    x = pdist(snitz_features.values, "cosine")
    snitz_distances = pd.DataFrame(
        squareform(x), index=snitz_features.index, columns=snitz_features.index
    )
    return snitz_distances



import csv
import os
import pathlib

import pyrfume
from pyrfume import Component, Mixture, Result, TriangleTest

# Parameters for Bushdid et al, 2014.
DILUTION = {"1/4": 0.25, "1/2": 0.5, "not diluted": 1.0}
CORRECT = {"right": True, "wrong": False}
OVERLAPS = {10: [9, 6, 3, 0], 20: [19, 15, 10, 5, 0], 30: [29, 20, 10, 0]}
N_TESTS = 20
N_SUBJECTS = 26
C = 128
BUSHDID_PATH = pathlib.Path("bushdid_2014")


def load_data():
    path = BUSHDID_PATH / "Bushdid-tableS1.csv"
    df = pyrfume.load_data(path, encoding="latin1")
    df = df.iloc[:, :4]
    df.columns = ["Name", "CAS", "Dilution", "Solvent"]
    df = df.set_index("CAS")
    df = df[df.index.notnull()]
    df.head()
    return df


# Functions for loading data from Bushdid et al, 2014.
def load_odorants_tests_results(all_components):
    """
    Given all odor components, loads the odorants, tests, and test results
    from Supplemental Table 2 of Bushdid et al.
    """
    odorants = {}
    tests = {}
    results = []
    path = pyrfume.DATA_DIR / BUSHDID_PATH / "Bushdid-tableS2.csv"
    f = open(path, "r", encoding="latin1")
    reader = csv.reader(f)
    next(reader)
    row_num = 0
    for row in reader:
        uid, n, r, percent, dilution, correct = row[:6]
        component_names = [x for x in row[6:36] if len(x)]
        # The next line is required to account for inconsistent naming of one
        # of the components across the two supplemental tables.
        component_names = [
            x.replace("4-Methyl-3-penten-2-one", "4-methylpent-3-en-2-one") for x in component_names
        ]
        outcomes = row[36:62]
        if uid.isdigit():
            uid = int(uid)
            dilution = DILUTION[dilution]
            odorant_key = hash(tuple(component_names))
            if odorant_key not in odorants:
                components = [
                    component for component in all_components if component.name in component_names
                ]
                if len(components) not in [1, 10, 20, 30]:
                    # If an odorant has a number of components which is not
                    # either 1, 10, 20, or 30.
                    print(
                        uid, [x for x in component_names if x not in [y.name for y in components]]
                    )
                odorant = Mixture(C, components)
                odorant.name = odorant_key
            elif row_num % 3 == 0:
                # If any component is repeated across all the tests.
                print("Repeat of this odorant: %d" % odorant_key)
            odorants[odorant_key] = odorant
            if uid not in tests:
                tests[uid] = TriangleTest(uid, [], dilution, correct)
            test = tests[uid]
            test.add_odorant(odorant)
            if correct == "right":
                test.correct = tests[uid].odorants.index(odorant)
        if len(outcomes[0]):
            for i, outcome in enumerate(outcomes):
                result = Result(test, i + 1, CORRECT[outcome])
                results.append(result)
        row_num += 1
    return odorants, tests, results


def load_components():
    """
    Loads all odorant components from Supplemental Table 1 of Bushdid et al.
    """

    components = []
    path = os.path.join(BUSHDID_PATH, "Bushdid-tableS1.csv")
    f = open(path, "r", encoding="latin1")
    reader = csv.reader(f)
    next(reader)
    component_id = 0
    for row in reader:
        name, cas, percent, solvent = row[:4]
        if len(name):
            component = Component(component_id, name, cas, percent, solvent)
            components.append(component)
            component_id += 1
        else:
            break
    return components


def get_results():
    components = load_components()
    odorants, tests, results = load_odorants_tests_results(components)
    return results



import json
import numpy as np
import urllib


class Mixture(object):
    """
    A mixture of molecules, defined by the presence of absence of the
    candidate molecules in the mixture.
    """

    def __init__(self, C: int, components: list=None):
        """
        Builds odorant from a list of components.
        """
        self.C = C
        self.components = components if components else []

    name = None  # Name of odorant, built from a hash of component names.

    C = None  # Number of components from which to choose.

    def components_vector(self, all_components: list=None, normalize: float=0):

        vector = np.zeros(self.C)
        for component in self.components:
            vector[all_components.index(component)] = 1
        if normalize:
            denom = (np.abs(vector) ** normalize).sum()
            vector /= denom
        return vector

    @property
    def N(self):
        """
        Number of components in this odorant.
        """

        return len(self.components)

    def r(self, other):
        """
        Number of replacements (swaps) to get from self to another odorant.
        """

        if len(self.components) == len(other.components):
            return self.hamming(other) / 2
        else:
            return None

    def overlap(self, other, percent=False):
        """
        Overlap between self and another odorant.  Complement of r.
        Optionally report result as percent relative to number of components.
        """

        overlap = self.N - self.r(other)
        if percent:
            overlap = overlap * 100.0 / self.N
        return overlap

    def hamming(self, other):
        """
        Hamming distance between self and another odorant.
        Synonymous with number of d, the number of total 'moves' to go from
        one odorant to another.
        """

        x = set(self.components)
        y = set(other.components)
        diff = len(x) + len(y) - 2 * len(x.intersection(y))
        return diff

    def add_component(self, component):
        """
        Adds one component to an odorant.
        """

        self.components.append(component)

    def remove_component(self, component):
        """
        Removes one component to an odorant.
        """

        self.components.remove(component)

    def descriptor_list(self, source):
        """
        Given a data source, returns a list of descriptors about this odorant.
        """

        descriptors = []
        for component in self.components:
            if source in component.descriptors:
                desc = component.descriptors[source]
                if type(desc) == list:
                    descriptors += desc
                if type(desc) == dict:
                    descriptors += [key for key, value in list(desc.items()) if value > 0.0]
        return list(set(descriptors))  # Remove duplicates.

    def descriptor_vector(self, source, all_descriptors):
        """
        Given a data source, returns a vector of descriptors about this
        odorant. The vector will contain positive floats.
        """

        vector = np.zeros(len(all_descriptors[source]))
        for component in self.components:
            if source in component.descriptors:
                desc = component.descriptors[source]
                if type(desc) == list:
                    for descriptor in desc:
                        index = all_descriptors[source].index(descriptor)
                        assert index >= 0
                        vector[index] += 1
                if type(desc) == dict:
                    this_vector = np.array([value for key, value in sorted(desc.items())])
                    vector += this_vector
        return vector

    def descriptor_vector2(self, all_descriptors):
        """
        Returns a vector of descriptors about this odorant, combining multiple
        data sources.
        """

        n_descriptors_dravnieks = len(all_descriptors["dravnieks"])
        n_descriptors_sigma_ff = len(all_descriptors["sigma_ff"])
        vector = np.zeros(n_descriptors_dravnieks + n_descriptors_sigma_ff)
        for component in self.components:
            if "dravnieks" in component.descriptors:
                desc = component.descriptors["dravnieks"]
                this_vector = np.array([value for key, value in sorted(desc.items())])
                vector[0:n_descriptors_dravnieks] += this_vector
            elif "sigma_ff" in component.descriptors:
                desc = component.descriptors["sigma_ff"]
                for descriptor in desc:
                    index = all_descriptors["sigma_ff"].index(descriptor)
                    assert index >= 0
                    vector[n_descriptors_dravnieks + index] += 1
        return vector

    def described_components(self, source):
        """
        Given a data source, returns a list of the components which are
        described by that source, i.e. those that have descriptors.
        """

        return [component for component in self.components if source in component.descriptors]

    def n_described_components(self, source):
        """
        Given a data source, returns the number of components that are
        described by that data source.
        """

        return len(self.described_components(source))

    def fraction_components_described(self, source):
        """
        Given a data source, returns the fraction of components that are
        described by that data source.
        """

        return self.n_described_components(source) / self.N

    def matrix(self, features, weights=None):
        matrix = np.vstack(
            [
                component.vector(features, weights=weights)
                for component in self.components
                if component.cid in features
            ]
        )
        if 0:  # matrix.shape[0] != self.N:
            print(
                ("Mixture has %d components but only " "%d vectors were computed")
                % (self.N, matrix.shape[0])
            )
        return matrix

    def vector(self, features, weights=None, method="sum"):
        matrix = self.matrix(features, weights=weights)
        if method == "sum":
            vector = matrix.sum(axis=0)
        else:
            vector = None
        return vector

    def __str__(self):
        """
        String representation of the odorant.
        """

        return ",".join([str(x) for x in self.components])


class Component(object):
    """
    A single molecule, which may or may not be present in an odorant.
    """

    def __init__(self, component_id: int, name: str, cas: str, percent: float, solvent: "Compound"):
        """
        Components are defined by a component_id from the Bushdid et al
        supplemental material, a name, a CAS number, a percent dilution,
        and a solvent.
        """

        self.id = component_id
        self.name = name
        self.cas = cas
        self.cid_ = None
        self.percent = percent
        self.solvent = solvent
        self.descriptors = {}  # An empty dictionary.

    @property
    def cid(self):
        cid = None
        if self.cid_:
            cid = self.cid_
        else:
            url_template = "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/name/%s/cids/JSON"
            for query in self.cas, self.name:
                try:
                    url = url_template % query
                    page = urllib.request.urlopen(url)
                    string = page.read().decode("utf-8")
                    json_data = json.loads(string)
                    cid = json_data["IdentifierList"]["CID"][0]
                except urllib.error.HTTPError:
                    print(query)
                else:
                    break
            self.cid_ = cid
        return cid

    def set_descriptors(self, source: str, cas_descriptors) -> None:
        """
        Given a data source, sets descriptors for this odorant using
        a dictionary where CAS numbers are keys, and descriptors are values.
        """

        assert type(source) == str and len(source)
        if self.cas in cas_descriptors:
            self.descriptors[source] = cas_descriptors[self.cas]
            # For sigma_ff this will be a list.
            # For dravnieks this will be a dict.

    def vector(self, features, weights=None):
        if self.cid in features:
            feature_values = np.array(list(features[self.cid].values()))
            if weights is None:
                weights = np.ones(feature_values.shape)
            result = feature_values * weights
        else:
            result = None
        return result

    def __str__(self):
        return self.name



import pathlib

import pyrfume

MAINLAND_INTENSITY_PATH = pathlib.Path("mainland_intensity")


def load_mainland_data(extra_cols=[]):
    cols = ["Subject", "Odor", "CAS", "Concentration", "IntensityRating"]
    cols += extra_cols
    df = pyrfume.load_data(MAINLAND_INTENSITY_PATH / "all data Supra clean.csv", index_col=None)
    df = df[cols]
    df["Subject"] = df["Subject"].astype(int)
    df["IntensityRating"] /= 100
    return df


def load_threshold_compilation():
    """Load data from http://www.thresholdcompilation.com/paginas/odour/odour.html
    A license or purchased copy of the book is required to access this data"""
    pyrfume.load_data("threshold/parsed_threshold_data_in_air.csv")



import contextlib
import io
import math
import random
import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import dask.bag as db
from deap import algorithms, base, creator, tools


@contextlib.contextmanager
def suppress_stdout(on):
    if on:
        sys.stdout = io.StringIO()
    yield
    sys.stdout = sys.__stdout__


class OdorantSetOptimizer:
    def __init__(
        self,
        library: pd.DataFrame,
        n_desired: int,
        weights: list,
        keep_cids: bool = None,
        fitness=None,
        n_gen: int = 300,
        mu: int = 100,
        lamda: int = 200,
        p_cx: float = 0.4,
        p_mut: float = 0.4,
        sel="selBest",
        standardize_weights=False,
        npartitions=1,
    ):
        """
        Args:
            library (pd.DataFrame): A pandas dataframe containing odorants and
                their attributes. The index should contain chemical identifiers
                (e.g. CIDs). The column names should be attributes, e.g.
                molecular weight, price, etc.  They can also be descriptors.
            n_desired (int): The size of the set of odorants desired from this
                library.
            keep_cids (list): An iterable of CIDs that must exist in every
                candidate set. Obviously this must be less than `n_desired`.
            weights (list): A list of 3-tuples each containing the name of a
                weight, a method to apply to the library dataframe, and a
                weight to apply to the function's output.
            fitness: A fitness class from DEAP
            mu (int): Size of the population (number of odorant sets under
                consideration in each generation).
            lamda (int): ...
            p_cx (float): Crossover probability in each generation.
            p_mut (float): Mutation probability in each generation.
            sel: Selection algorithm, e.g. 'selBest' or 'selNSGA2'
            standardize_weights (bool): Interpret weights as multiples of stdev
                of a weight function
        """

        # Turn all constructor arguments into attributes
        for key, value in locals().items():
            setattr(self, key, value)

        if self.fitness is None:
            self.fitness = BetterFitness

        weight_names, weight_functions, weight_values = zip(*self.weights)
        assert len(weight_names) == len(set(weight_names)), "Weight names must all be unique."

        # Number of available odorants
        self.library_size = self.library.shape[0]

        if self.keep_cids:
            # Get library integer indices of the CIDs to keep
            is_kept = self.library.index.isin(self.keep_cids)
            self.keep = set(np.flatnonzero(is_kept))
            self.n_needed = self.n_desired - len(self.keep)

        # Setup DEAP fundamentals
        creator.create("Fitness", self.fitness, weights=weight_values)
        creator.create("Individual", set, fitness=creator.Fitness)
        self.toolbox = base.Toolbox()

        # Describe the process for selecting a random set of items
        self.toolbox.register("random_set", random.sample, range(self.library_size), self.n_desired)

        # Describe the process for creating a single set
        self.toolbox.register(
            "individual", tools.initIterate, creator.Individual, self.toolbox.random_set
        )

        # Says that the population is just a list of individuals
        self.toolbox.register("population", tools.initRepeat, list, self.toolbox.individual)
        self.toolbox.register("evaluate", self.eval_individual)
        self.toolbox.register("mate", self.crossover)
        self.toolbox.register("mutate", self.mutate)
        selection_algorithm = getattr(tools, sel)
        self.toolbox.register("select", selection_algorithm)

        # Potentially use Dask for multithreading
        if npartitions > 1:
            self.toolbox.register("map", self.dask_map)

        # Potentially standardize weights and scores
        if self.standardize_weights:
            self.compute_weight_stats()

    def compute_weight_stats(self):
        """Compute scores for a bunch of random individuals to use for
        standardizing fitness scores and thus making weights interpretable
        as # of sd's better or worse than random individuals"""
        n_iter = 100  # Number of random individuals to generate
        weight_names = [x[0] for x in self.weights]
        fitnesses = pd.DataFrame(index=range(n_iter), columns=weight_names)
        for i in range(n_iter):
            ind = np.random.choice(range(self.library_size), self.n_desired, replace=False)
            fitnesses.loc[i, :] = self.eval_individual(ind, standardize=False)
        self.stds = fitnesses.std()
        self.means = fitnesses.mean()

    def eval_individual(self, individual, standardize: bool = None):
        """Evaluate the fitness of the odorant set.

        Args:
            individual: A DEAP individual as registered above.
            standardize (bool, optional): Whether to standardize the
                resulting score.

        Returns:
            A fitness score.
        """
        fitness = []
        for weight_name, func_name, value in self.weights:
            if isinstance(func_name, str):
                f = getattr(self, "eval_%s" % func_name)
                component = f(individual, weight_name)
            elif isinstance(func_name, (tuple, list, set)):
                f, *args = func_name
                component = f(individual, *args)
            else:
                f = func_name
                component = f(individual)
            if self.standardize_weights and (standardize is not False):
                m = self.means[weight_name]
                s = self.stds[weight_name]
                component = (component - m) / s
            fitness.append(component)
        return fitness

    def crossover(self, ind1, ind2) -> tuple:
        """Apply a crossover operation on input sets.

        Args:
            ind1: The first individual
            ind2: The second individual

        Returns:
            tuple: Two new individuals
        """
        union = ind1 | ind2
        x1 = random.sample(union, self.n_desired)
        x2 = random.sample(union, self.n_desired)
        ind1.clear()
        ind1.update(x1)
        ind2.clear()
        ind2.update(x2)
        assert len(ind1) == self.n_desired, "Individual smaller than desired size."
        assert len(ind2) == self.n_desired, "Individual smaller than desired size."
        if self.keep_cids is not None:
            ind1 = self.force_keep(ind1)
            ind2 = self.force_keep(ind2)
        return ind1, ind2

    def mutate(self, individual) -> tuple:
        """Mutation that pops or adds some elements.

        Args:
            individual: An individual

        Returns:
            tuple: A new individual (wrapped in a tuple)
        """
        N_MUTATIONS = 1
        available = list(set(range(self.library_size)).difference(individual))
        to_remove = random.sample(list(individual), N_MUTATIONS)
        to_add = random.sample(available, N_MUTATIONS)
        # Removes elements found in `to_remove`
        # XOR works here to remove `to_remove` because it can
        # only contain current members of the individial
        individual ^= set(to_remove)
        # Adds elements found in `to_add`
        individual |= set(to_add)
        assert len(individual) == self.n_desired, "Individual smaller than desired size."
        if self.keep_cids is not None:
            individual = self.force_keep(individual)
        return (individual,)

    def force_keep(self, ind):
        """Ensure that the CIDs in `self.keep` get included

        Args:
            ind: An individual.

        Returns:
            An individual with certain additional items kept.
        """
        # Consider only molecules for this individual
        # which are not in the keep set
        candidates = ind.difference(self.keep)
        # Keep only enough of those that there are room for
        to_add = set(random.sample(candidates, self.n_needed))
        # Set this individual to contain this subset
        # of the candidates plus the keep set
        ind.clear()
        ind.update(to_add | self.keep)
        return ind

    def run(self, pop=None, hof=None, quiet: bool = False) -> tuple:
        """Run the optimization

        Args:
            pop: The population to start with.
            hof: The Hall of Fame to start with
            quiet: (bool, optional): Whether to suppress stdout.
                Defaults to False.

        Returns:
            tuple: The population, stats, Hall of Fame, and logboox
        """
        self.pop = pop if pop else self.toolbox.population(n=self.mu)
        if hof is None:
            self.hof = tools.HallOfFame(self.n_gen)
        self.stats = tools.Statistics(lambda ind: ind.fitness.values)
        self.stats.register("avg", lambda x: np.mean(x, axis=0))

        def best(x):
            weight_names, weight_functions, weight_values = zip(*self.weights)
            scores = [np.dot(self.hof.keys[i].values, weight_values) for i in range(len(self.hof))]
            i = np.argmax(scores)
            return np.array([scores[i]] + list(self.hof.keys[i].values))

        self.stats.register("best", best)

        f = algorithms.eaMuPlusLambda
        with suppress_stdout(quiet):
            self.pop, self.logbook = f(
                self.pop,
                self.toolbox,
                self.mu,
                self.lamda,
                self.p_cx,
                self.p_mut,
                self.n_gen,
                self.stats,
                halloffame=self.hof,
            )

        return self.pop, self.stats, self.hof, self.logbook

    def eval_mean(self, individual, column: str) -> float:
        """Evaluation the mean of column for the selected individual

        Args:
            individual: An individual.
            column (str): A column name from the library dataframe.

        Returns:
            float: The mean of that column, for the subset
                described by the individual
        """
        return self.library.iloc[list(individual)][column].mean()

    def eval_sum(self, individual, column: str):
        """Evaluation the sum of column for the selected individual

        Args:
            individual: An individual.
            column (str): A column name from the library dataframe.

        Returns:
            float: The sum of that column, for the subset
                described by the individual
        """
        return self.library.iloc[list(individual)][column].sum()

    def dask_map(self, f, x: list) -> list:
        """A mapping function for Dask.  Used for multithreading.

        Args:
            f: Any function.
            x (list): A list of inputs to be sequentially passed
                to that function.

        Returns:
            list: A list of outputs from that function.
        """
        x = db.from_sequence(x, npartitions=self.npartitions)
        return db.map(f, x).compute()

    def plot_score_history(self):
        """Plot the history of each of the scores over generations."""
        rows = 2
        cols = math.ceil((len(self.weights) + 1) / 2)
        fig, axes = plt.subplots(rows, cols, figsize=(20, 8))
        scores = [x["best"][0] for x in self.logbook]
        axes[0, 0].plot(scores)
        axes[0, 0].set_title("Total")
        for j, (feature, stat, weight) in enumerate(self.weights):
            ax = axes.flat[j + 1]
            history = [x["best"][j + 1] for x in self.logbook]
            ax.plot(history)
            ax.set_title(feature)
        axes[0, 0].set_xlabel("Generation")
        plt.tight_layout()


class BetterFitness(base.Fitness):
    def __le__(self, other):
        return sum(self.wvalues) <= sum(other.wvalues)

    def __lt__(self, other):
        return sum(self.wvalues) < sum(other.wvalues)


def get_coverage(odorant_indices, space, sigma=2):
    """This function will be used to compute coverage of odorant space during
    optimization. We want non-selected odorants to be 'covered' as much as
    possible by selected ones"""
    ind = list(odorant_indices)
    return space.iloc[ind].max(axis=0).mean()


def get_entropy(odorant_indices, space, bins_per_dim=10):
    from scipy.stats import entropy

    """These function will be used to determine entropy of the selected
    odorants in the low-d manifold. We want the selected odorants to smoothly
    cover the available odorant space"""
    ind = list(odorant_indices)
    # Determine optimal bins in each dimension
    bins = {dim: np.histogram_bin_edges(space[dim], bins=bins_per_dim) for dim in ("X", "Y")}
    # Get the XY coordinates of the selected odorants in the embedding
    XY = space.iloc[ind].values
    # Compute a 2D histogram of selected odorant counts in each bin
    histXY = np.histogram2d(*np.array(XY).T, bins=(bins["X"], bins["Y"]))[0]
    # Compute the entropy of this histogram
    # Higher is flatter (less clustered)
    h = entropy(histXY.ravel())
    return h


def get_spacing(odorant_indices, space, n=5):
    """These function will be used to determine the spacing between the
    selected odorants in the low-d manifold. We don't want to select odorants
    which are right next to each other in odorant space"""
    ind = list(odorant_indices)
    # Pairwise distances between selected odorants
    x = space.iloc[ind, ind].values
    # Unraveled (ignoring diagonal)
    x = x[np.triu_indices(len(ind), 1)]
    # Penalize the closest n odorant pairs
    return np.sort(x)[:n].mean()


# Example behavior:
# np.set_printoptions(precision=2, suppress=True)
# library = pd.read_csv('path/to/odorants.csv')#
# weights = [('col1', 'mean', 1),
#            ('col2', 'sum', 2),
#            ('name3', (func, arg1, arg2), -3)]
# optimizer = OdorantSetOptimizer(library, 25, weights, n_gen=25)
# pop, stats, hof, logbook = optimizer.run()
# library.iloc[list(hof[0])]



from math import exp
from math import log as ln

import quantities as pq
from sympy import Symbol
from sympy.solvers import solve


def mackay(vp):
    """Mackay, D., & van Wesenbeeck, I. (2014).
    Correlation of Chemical Evaporation Rate with Vapor Pressure.
    Environmental Science & Technology, 48(17), 10259–10263.
    doi:10.1021/es5029074
    Note typo in intercept parameter in Figure 1.
    """
    # Units of Pascals
    vp = vp.rescale(pq.Pa)
    # Strip units for logarithm
    vp /= vp.units
    # # Evaporation rate
    if vp > 0:
        er = exp(1.0243 * ln(vp) - 15.08)
    else:
        er = 0
    # Attach units
    er *= pq.mol / (pq.m ** 2 * pq.s)
    return er


def bernoulli(v=None, p=None, rho=None, g=9.8 * pq.m / (pq.s) ** 2, z=0, k=None):
    g = float(g.simplified)
    if v is None:
        v = Symbol("v", real=True, positive=True)
    else:
        v = v.rescale(pq.m / pq.s)
        v = float(v.simplified)
    if p is None:
        p = Symbol("p", real=True, positive=True)
    else:
        p = p.rescale(pq.Pa)
        p = float(p.simplified)
    if rho is None:
        rho = Symbol("rho", real=True, positive=True)
    else:
        rho = rho.rescale(pq.kg / (pq.m ** 3))
        rho = float(rho.simplified)
    if k is None:
        k = Symbol("k", real=True, positive=True)
    else:
        k = k.rescale((pq.m / pq.s) ** 2)
        k = float(k.simplified)
    result = solve((v ** 2) / 2 + g * z + p / rho - k, [v, p, rho], dict=True)
    return result


def venturi(rho=None, p1=None, p2=None, v1=None, v2=None):
    assert rho is not None
    rho = rho.rescale(pq.kg / (pq.m ** 3))
    rho = float(rho.simplified)

    if v1 is None:
        v1 = Symbol("v1", real=True, positive=True)
    else:
        v1 = v1.rescale(pq.m / pq.s)
        v1 = float(v1.simplified)
    if v2 is None:
        v2 = Symbol("v2", real=True, positive=True)
    else:
        v2 = v2.rescale(pq.m / pq.s)
        v2 = float(v2.simplified)
    if p1 is None:
        p1 = Symbol("p1", real=True, positive=True)
    else:
        p1 = p1.rescale(pq.Pa)
        p1 = float(p1.simplified)
    if p2 is None:
        p2 = Symbol("p2", real=True, positive=True)
    else:
        p2 = p2.rescale(pq.Pa)
        p2 = float(p2.simplified)

    result = solve(rho * (v2 ** 2 - v1 ** 2) / 2 - p1 + p2, [v1, v2, p1, p2], dict=True)
    return result


if __name__ == "__main__":
    # print(bernoulli(None, 10*pq.psi, 1.225*pq.kg/(pq.m**3)))
    print(
        venturi(rho=1.225 * pq.kg / (pq.m ** 3), p1=10 * pq.psi, p2=1 * pq.psi, v1=10 * pq.m / pq.s)
    )



from nbconvert.preprocessors import ExecutePreprocessor
import nbformat
from pathlib import Path
import platform
import unittest


class DocsTest(unittest.TestCase):
    def setUp(self):
        import pyrfume
        pyrfume_dir = Path(pyrfume.__path__[0]).parent
        self.docs_path = pyrfume_dir / 'docs'
        #pyrfume.set_data_path(pyrfume_dir.parent / 'pyrfume-data')
    
    def execute(self, path):
        with open(path) as f:
            nb = nbformat.read(f, as_version=4)
        ep = ExecutePreprocessor(timeout=600, kernel_name='python3')
        ep.preprocess(nb)
        
    def execute_doc(self, name):
        self.execute(self.docs_path / (name+'.ipynb'))

    def test_1_your_data(self):
        self.execute_doc('your-data')
        
    @unittest.skipIf('spike' not in platform.node(), "Only tested on Spike")
    def test_2_featurization(self):
        self.execute_doc('featurization')
        
    def test_3_visualization(self):
        self.execute_doc('visualization')
        
    def test_4_published_data(self):
        self.execute_doc('published-data')


if __name__ == '__main__':
    unittest.main()



import pyrfume

from . import dream


def load_raw_bmc_data(nrows=None):
    """Load raw data from Keller and Vosshall, 2016 supplement."""
    df_raw = pyrfume.load_data("keller_2016/12868_2016_287_MOESM1_ESM.xlsx", header=2)
    return df_raw


def format_bmc_data(
    df,  # The raw data frame returned by `load_raw_bmc_data`
    only_dream_subjects=False,  # Whether to only keep DREAM subjects
    only_dream_descriptors=False,  # Whether to only keep DREAM descriptors
    only_dream_molecules=False,
):  # Whether to only keep DREAM molecules
    """Format raw data from the BMC paper to be usable for modeling"""
    # Remove leading and trailing white space from column names
    df.columns = df.columns.str.strip()

    # Get the raw DREAM descriptor list
    descriptors_raw = dream.get_descriptors()
    # Get the publication-style descriptor names
    descriptors = dream.get_descriptors(format=True)
    # Revise to the Keller and Vosshall descriptor names
    descriptors_raw[0] = "HOW STRONG IS THE SMELL?"
    descriptors_raw[1] = "HOW PLEASANT IS THE SMELL?"

    # Possibly include "Familiarity" as a descriptor
    if not only_dream_descriptors:
        descriptors_raw.append("HOW FAMILIAR IS THE SMELL?")
        descriptors.append("Familiarity")

    # Possibly restrict subjects to those used in the DREAM challenge
    # Note that numeric subject IDs in the BMC paper and in the DREAM
    # challenge are not identical
    if only_dream_subjects:
        df["Subject"] = df["Subject # (DREAM challenge)"].fillna(0).astype(int)
        df = df[df["Subject"] > 0]
    else:
        df["Subject"] = df["Subject # (this study)"].astype(int)

    # Rename columns to match DREAM challenge
    df = df.rename(columns={"Odor dilution": "Dilution"})
    df = df.rename(columns=dict(zip(descriptors_raw, descriptors)))

    # Fix CIDs for molecules that only have CAS registry numbers.
    # Geranylacetone didn't have a CID listed in the raw data
    # Isobutyl acetate had the wrong CAS number in the raw data
    df["CID"] = (
        df["CID"]
        .astype(str)
        .str.replace("3796-70-1", "1549778")
        .str.replace("109-19-0", "8038")
        .astype(int)
    )

    # Possibly keep only the 476 DREAM challenge molecules
    if only_dream_molecules:
        dream_CIDs = dream.get_cids()
        assert len(dream_CIDs) == 476
        df = df[df["CID"].isin(dream_CIDs)]

    # Keep only relevant columns
    df = df[["CID", "Dilution", "Subject"] + descriptors]

    # Fill NaN descriptors values with 0 if Intensity is not 0.
    df = df.apply(lambda x: x.fillna(0) if x["Intensity"] > 0 else x, axis=1)

    # Make dilution values integer -log10 dilutions
    df["Dilution"] = df["Dilution"].apply(dream.dilution_to_magnitude).astype(float)

    # Set index and set column axis name
    df = df.set_index(["CID", "Dilution", "Subject"])
    df.columns.name = "Descriptor"

    # Identify replicates and add this information to the index
    df["Replicate"] = df.index.duplicated().astype(int)
    df = df.reset_index().set_index(["CID", "Dilution", "Replicate", "Subject"])
    if only_dream_subjects:
        # DREAM subjects replicates should be properly indexed now
        assert df.index.duplicated().sum() == 0

    # Rearrange dataframe to pivot subjects and descriptors
    df = df.unstack("Subject").stack("Descriptor")
    df = df.reorder_levels(["Descriptor", "CID", "Dilution", "Replicate"])
    df = df.sort_index()

    return df



import unittest
import quantities as pq
from datetime import datetime
from pyrfume.objects import Mixture, Component
from pyrfume.odorants import Compound, ChemicalOrder, Vendor, Molecule

class MixtureTestCase(unittest.TestCase):
    def test_constructor(self):
        mixture1 = Mixture(0)
        mixture2 = Mixture(9999)
        self.assertEqual(mixture1.N, 0)
        self.assertEqual(mixture1.r(mixture2), 0)
        self.assertEqual(mixture1.overlap(mixture2), 0)
        self.assertEqual(mixture1.hamming(mixture2), 0)

    def test_example(self):
        cid = 21946271
        cas = "12002-61-8"
        vendor = Vendor("Test_Vendor", "https://en.wikipedia.org/wiki/Actinium(III)_oxide")
        molecule = Molecule(cid, "test_mol", False)
        molecule = Molecule(cid, "test_mol", True)
        self.assertIsNone(molecule.molarity)
        molecule.molecular_weight = pq.Quantity(502.1, pq.g / pq.mol)
        molecule.density = pq.Quantity(501.9982, pq.g / pq.cm ** 3)
        self.assertAlmostEqual(molecule.molarity.item(), 999.797251, 3)

        chemical_order = ChemicalOrder(molecule, vendor, "", 0.9, None)
        compound = Compound(chemical_order, "TEST", datetime.now, datetime.now, False)

        self.assertIsInstance(getattr(compound, "vendor"), Vendor)
        self.assertIsInstance(getattr(compound, "cid"), int)

        component = Component(cid, "name1", cas, 0.5, compound)
        component1 = Component(cid, "name2", cas, 0.8, compound)
        mixture = Mixture(1, [component])

        components_vector = mixture.components_vector(mixture.components, 1)
        self.assertEqual(components_vector.size, 1)

        mixture1 = Mixture(0)
        self.assertIsNone(mixture1.r(mixture))
        self.assertEqual(mixture.overlap(mixture, True), 100.0)
        
        mixture.add_component(component1)
        mixture.remove_component(component1)
        self.assertEqual(len(mixture.components), 1)
        mixture.remove_component(component)
        self.assertEqual(len(mixture.components), 0)

        cas_descriptor = {
            cas : ["cas descriptor"],
            "dravnieks" : ["dravnieks descriptor"],
            "sigma_ff" : ["sigma_ff descriptor"]
        }
        cas_descriptor1 = {
            cas : {"cas descriptor 1" : 1},
            "dravnieks" : {"dravnieks cas description 1" : 2},
            "sigma_ff" : {"sigma_ff descriptor 1"}
        }
        component.set_descriptors("unittest source", cas_descriptor)
        component1.set_descriptors("unittest source", cas_descriptor1)
        mixture.add_component(component)
        mixture.add_component(component1)

        descriptors_list = mixture.descriptor_list("unittest source")
        self.assertTrue('cas descriptor 1' in descriptors_list)
        self.assertTrue('cas descriptor' in descriptors_list)

        all_descriptors = {
            'unittest source' : 'cas descriptor 1', 
            'unittest source' : 'cas descriptor'
        }
        self.assertEqual(mixture.descriptor_vector("unittest source", all_descriptors).size, 14)
        
        all_descriptors = {
            'dravnieks' : "dravnieks cas description 1", 
            'sigma_ff' : "sigma_ff descriptor 1"
        }

        self.assertEqual(mixture.descriptor_vector2(all_descriptors).size, 48)

        described_components = mixture.described_components("unittest source")
        self.assertTrue(component in described_components)
        self.assertTrue(component1 in described_components)

        n = mixture.n_described_components("unittest source")
        self.assertEqual(n, 2)

        self.assertEqual(mixture.fraction_components_described("unittest source"), 1)

        test_feature = {cid : {"a" : 1}}
        self.assertEqual(mixture.matrix(test_feature).shape, (2, 1))
        self.assertEqual(mixture.vector(test_feature)[0], 2.0)
        self.assertIsInstance(str(mixture), str)
        self.assertIsInstance(str(mixture1), str)

        self.assertEqual(component.cid, 21946271)
        self.assertIsInstance(str(component), str)
        self.assertIsInstance(str(component1), str)

if __name__ == '__main__':
    unittest.main()



import unittest

from pyrfume.odorants import Compound, ChemicalOrder, Vendor, Molecule
from pyrfume.experiments import TriangleTest, Result
from pyrfume.objects import Component, Mixture
from datetime import datetime

from .unittest_utils import get_substances

class TriangleTestTestCase(unittest.TestCase):

    def test_test(self):

        mixture_C4H8O2, mixture_C4H8S, mixture_C2H6O = get_substances("mixtures")
        component_C4H8O2, component_C4H8S, component_C2H6O = get_substances("components")

        odorants = [mixture_C4H8O2]

        test = TriangleTest(0, odorants, 1.0, True)
        test.add_odorant(mixture_C2H6O)
        test.add_odorant(mixture_C2H6O)

        self.assertEqual(mixture_C2H6O, test.double)
        self.assertEqual(mixture_C4H8O2, test.single)
        self.assertEqual(mixture_C2H6O, test.pair[0])
        self.assertEqual(mixture_C4H8O2, test.pair[1])
        

        odorants = [mixture_C4H8O2]

        test = TriangleTest(0, odorants, 1.0, True)
        test.add_odorants([mixture_C2H6O, mixture_C2H6O])
        
        self.assertEqual(mixture_C2H6O, test.double)
        self.assertEqual(mixture_C4H8O2, test.single)
        self.assertEqual(mixture_C2H6O, test.pair[0])
        self.assertEqual(mixture_C4H8O2, test.pair[1])

        self.assertEqual(1, test.N)
        self.assertEqual(1, test.r)
        self.assertEqual(0, test.overlap(False))
        self.assertEqual(0, test.overlap(True))        

        self.assertTrue("common descriptor" in test.common_descriptors('unittest source'))
        self.assertTrue("C2H6O unique descriptor" in test.unique_descriptors('unittest source'))
        self.assertTrue("C4H8O2 unique descriptor" in test.unique_descriptors('unittest source'))
        all_descriptors = {
            'unittest source' : [
                'C2H6O unique descriptor', 
                'C4H8O2 unique descriptor',
                "common descriptor"
            ],
            "dravnieks" : [
                "C2H6O dravnieks descriptor",
                "C4H8S dravnieks descriptor",
                "C4H8O2 dravnieks descriptor", 
                "common dravnieks descriptor"
            ],
            "sigma_ff" : [
                "C2H6O sigma_ff descriptor",
                "C4H8S sigma_ff descriptor",
                "C4H8O2 sigma_ff descriptor", 
                "common sigma_ff descriptor"
            ]
        }
        self.assertAlmostEqual(
            test.descriptors_correlation('unittest source', all_descriptors), -0.5, 2
        )

        self.assertAlmostEqual(
            test.descriptors_correlation('unittest source', all_descriptors), -0.5, 2
        )

        test.descriptors_correlation2(all_descriptors)
        diff_list = list(test.descriptors_difference('unittest source', all_descriptors))
        self.assertEqual(diff_list.count(1.0), 2)
        self.assertEqual(diff_list.count(0), 1)

        self.assertEqual(len(test.common_components), 0)
        mixture_C2H6O_C4H8S = mixture_C2H6O
        mixture_C2H6O_C4H8S.add_component(component_C4H8S)
        mixture_C4H8O2_C4H8S = mixture_C4H8O2
        mixture_C4H8O2_C4H8S.add_component(component_C4H8S)
        unique_components = test.unique_components
        unique_components = [str(x) for x in unique_components]
        self.assertTrue(component_C4H8S in test.common_components)
        self.assertTrue(str(component_C2H6O) in unique_components)
        self.assertTrue(str(component_C4H8O2) in unique_components)
        self.assertEqual(test.n_undescribed('unittest source'), (0, 0))

        result1 = Result(test, 0, True)
        results = [result1]
        self.assertEqual(test.fraction_correct(results), 1)

        test2 = TriangleTest(0, odorants, 1.0, True)
        result2 = Result(test2, 0, False)
        results.append(result2)
        self.assertEqual(test.fraction_correct(results), 0.5)

if __name__ == '__main__':
    unittest.main()



"""Utilities for DREAM challenge data"""

import numpy as np

import pyrfume


def get_descriptors(format=False):
    """Get DREAM challenge descriptors"""
    if format:
        result = [
            "Intensity",
            "Pleasantness",
            "Bakery",
            "Sweet",
            "Fruit",
            "Fish",
            "Garlic",
            "Spices",
            "Cold",
            "Sour",
            "Burnt",
            "Acid",
            "Warm",
            "Musky",
            "Sweaty",
            "Ammonia",
            "Decayed",
            "Wood",
            "Grass",
            "Flower",
            "Chemical",
        ]
    else:
        result = [
            "INTENSITY/STRENGTH",
            "VALENCE/PLEASANTNESS",
            "BAKERY",
            "SWEET",
            "FRUIT",
            "FISH",
            "GARLIC",
            "SPICES",
            "COLD",
            "SOUR",
            "BURNT",
            "ACID",
            "WARM",
            "MUSKY",
            "SWEATY",
            "AMMONIA/URINOUS",
            "DECAYED",
            "WOOD",
            "GRASS",
            "FLOWER",
            "CHEMICAL",
        ]
    return result


def get_cids():
    """Get PubChem IDs for DREAM challenge molecules"""
    df = pyrfume.load_data("keller_2017/cids.csv", index_col=None)
    return list(df["CID"])


def dilution_to_magnitude(dilution):
    denom = dilution.replace('"', "").replace("'", "").split("/")[1].replace(",", "")
    return np.log10(1.0 / float(denom))



from pathlib import Path
import unittest

class DataAndConfigTestCase(unittest.TestCase):

    def test_init_reset_config(self):
        from pyrfume import init_config, reset_config
        init_config(False)
        reset_config()
        init_config(True)
        
    def test_read_write_config(self):
        from pyrfume import read_config, write_config
        write_config("PATHS", "a", "b")
        self.assertEqual(read_config("PATHS", "a"), "b")

    def test_data_path(self):
        from pyrfume import set_data_path, get_data_path
        from pyrfume.base import PACKAGE_DIR, DEFAULT_DATA_PATH
        import os

        curr_data_path = get_data_path()
        
        path_not_exists = PACKAGE_DIR / "THIS_IS_AN_INVALID_PATH"
        self.assertRaises(Exception, set_data_path, path_not_exists, create=False)
        self.assertRaises(Exception, get_data_path, path_not_exists, create=False)

        path1 = PACKAGE_DIR / "unit_test"
        set_data_path(path1)
        path2 = get_data_path()
        self.assertEqual(path1, path2)

        Path(DEFAULT_DATA_PATH).mkdir(exist_ok=True)
        set_data_path(DEFAULT_DATA_PATH)
        path3 = get_data_path()
        self.assertEqual(path3, DEFAULT_DATA_PATH)

        set_data_path(curr_data_path)

    def test_load_data(self):
        import pickle, os
        from pyrfume.base import DEFAULT_DATA_PATH
        from pyrfume import load_data, save_data
        import pandas as pd

        data = {'col1': [1, 2], 'col2': [3, 4]}
        file_path = DEFAULT_DATA_PATH / "data.pkl"
        path_not_exists = DEFAULT_DATA_PATH / "THIS_IS_AN_INVALID_PATH"
        
        self.assertRaises(Exception, save_data, data, path_not_exists)
        save_data(data, file_path)

        data_gain = load_data(file_path)
        self.assertEqual(data_gain, data)
        os.remove(file_path)

        file_path = DEFAULT_DATA_PATH / "data.csv"

        df = pd.DataFrame(data)
        save_data(df, file_path)
        #with open(file_path, "w") as f:
        #    f.write("0,1,2,3\n0,1,2,3")

        data_gain = load_data(file_path)

        for index1 in range(len(data_gain.values)):
            for index2 in range(len(data_gain.values[index1])):
                self.assertEqual(data_gain.values[index1][index2], df.values[index1][index2])
                
        os.remove(file_path)

    def test_save_data(self):
        from pyrfume.base import DEFAULT_DATA_PATH


if __name__ == '__main__':
    unittest.main()



import warnings

import numpy as np
import pandas as pd
from sklearn.preprocessing import MinMaxScaler, Normalizer, StandardScaler

from eden.graph import Vectorizer
from pyrfume import load_data, save_data, logger, odorants
from pyrfume.mol2networx import smiles_to_eden
from tqdm.auto import tqdm, trange

try:
    from mordred import Calculator, descriptors as all_descriptors
    from rdkit import Chem
    from rdkit.Chem import AllChem, SaltRemover
    from rdkit import DataStructs
except (ImportError, ModuleNotFoundError):
    warnings.warn(
        "Parts of mordred and/or rdkit could not be imported; try installing rdkit via conda",
        UserWarning,
    )

FEATURES_DIR = "physicochemical"
DRAGON_STEM = "AllDragon%s.csv"


def load_dragon(suffix=""):
    """Loads dragon features.
    Use a suffix to specify a precomputed cleaning of this data"""
    path = "%s/%s" % (FEATURES_DIR, DRAGON_STEM % suffix)
    dragon = load_data(path)
    return dragon


def clean_features(raw_features, max_nan_frac=0.3):
    n_molecules = raw_features.shape[0]
    n_allowed_nans = n_molecules * max_nan_frac
    # Remove features with too many NaNs
    good = raw_features.columns[raw_features.isnull().sum() < n_allowed_nans]
    cleaned_features = raw_features[good]
    cols = [c for c in list(cleaned_features) if c != "SMILES"]
    cleaned_features = cleaned_features[cols]  # Drop SMILES column
    return cleaned_features


def scale_features(cleaned_features, scaler):
    if scaler == "standardize":
        # Scale to mean 0, variance 1
        s = StandardScaler()
    elif scaler == "normalize":
        # Scale to length 1
        s = Normalizer()
    elif scaler == "minmax":
        # Scale to min 0, max 1
        s = MinMaxScaler()
    else:
        raise Exception(("scaler must be one of 'standardize'," " 'normalize', or 'minmax'"))
    scaled_data = s.fit_transform(cleaned_features.astype("float"))
    scaled_features = pd.DataFrame(
        scaled_data, index=cleaned_features.index, columns=cleaned_features.columns
    )
    return scaled_features


def impute_features(scaled_features):
    # Impute missing values
    from fancyimpute import KNN

    knn = KNN(k=5)
    imputed_values = knn.fit_transform(scaled_features.values)
    imputed_features = pd.DataFrame(
        imputed_values, index=scaled_features.index, columns=scaled_features.columns
    )
    return imputed_features


def save_dragon(dragon, suffix):
    path = "%s/%s" % (FEATURES_DIR, DRAGON_STEM % suffix)
    save_data(dragon, path)


def cid_names():
    """TODO: Fix this to use the larger file"""
    path = FEATURES_DIR / "cids-names-smiles.csv"
    names = load_data(path).set_index("CID")["name"]
    return names


def mol_to_mordred(mols, features=None):
    calc = Calculator(all_descriptors)
    print("\nComputing Mordred features...")
    df = calc.pandas(mols.values())
    df = df.fill_missing()  # Use NaN instead of Missing object
    if features is not None:
        df = df[features]  # Retain only the specified features
    mordred = pd.DataFrame(df.values, index=mols.keys(), columns=df.columns)
    print("There are %d molecules and %d features" % mordred.shape)
    return mordred


def smiles_to_mordred(smiles, features=None, max_attempts=10):
    mols = odorants.smiles_to_mol(smiles, max_attempts=max_attempts)
    mordred = mol_to_mordred(mols, features=features)
    return mordred


def make_graphs(smiles):
    eden_graph_generator = [
        smiles_to_eden(smi) for smi in smiles
    ]  # Convert from SMILES to EdEN format
    graphs = [graph for graph in eden_graph_generator]  # Compute graphs for each molecule
    vectorizer = Vectorizer(min_r=0, min_d=0, r=1, d=2)
    sparse = vectorizer.transform(graphs)  # Compute the NSPDK features and store in a sparse array
    return sparse


def smiles_to_nspdk(smiles, features=None):
    nspdk_sparse = make_graphs(smiles)  # Compute the NSPDK features and store in a sparse array
    n_molecules, n_features = nspdk_sparse.shape
    print(
        "There are %d molecules and %d potential features per molecule" % (n_molecules, n_features)
    )
    # Extract the indices of NSPDK features where at least one molecules is non-zero
    if features is None:
        original_indices = sorted(list(set(nspdk_sparse.nonzero()[1])))
        n_used_features = len(original_indices)
        print(
            "Only %d of these features are used "
            "(%.1f features per molecule; %.1f molecules per feature)"
            % (
                n_used_features,
                nspdk_sparse.size / n_molecules,
                nspdk_sparse.size / n_used_features,
            )
        )
        # Create a dense array from those non-zero features
        nspdk_dense = nspdk_sparse[:, original_indices].todense()
        indices = original_indices
    else:
        n_used_features = len(features)
        nspdk_sparse = nspdk_sparse[:, features]
        print(
            "Only %d of these features will be used "
            "(%.1f features per molecule; %.1f molecules per feature)"
            % (
                n_used_features,
                nspdk_sparse.size / n_molecules,
                nspdk_sparse.size / n_used_features,
            )
        )
        nspdk_dense = nspdk_sparse.todense()  # Include only the desired features
        indices = features
    # Create a Pandas DataFrame
    nspdk = pd.DataFrame(nspdk_dense, index=smiles, columns=indices)
    return nspdk


def smiles_to_nspdk_sim(smiles, ref_smiles, features=None):
    sparse = make_graphs(smiles)
    ref_sparse = make_graphs(ref_smiles)
    sim = sparse.dot(ref_sparse.T).todense()
    sim = pd.DataFrame(sim, index=smiles, columns=ref_smiles)
    return sim


def smiles_to_dragon(smiles, suffix="", features=None):
    dragon = load_data("physicochemical/AllDragon%s.csv" % suffix)
    if features is None:
        features = list(dragon)
    dragon = dragon.loc[smiles, dragon]
    return dragon


def smiles_to_morgan(smiles, radius=5, features=None):
    mols = odorants.smiles_to_mol(smiles)
    morgan = mol_to_morgan(mols, radius=radius, features=features)
    return morgan


def mol_to_morgan(mols, radius=5, features=None):
    fps = []
    for index, mol in tqdm(mols.items(), desc='Computing Morgan Fingerprints'):
        fp = AllChem.GetMorganFingerprint(mol, radius)
        fp = pd.Series(fp.GetNonzeroElements(), name=index)
        fps.append(fp)
    
    batch_size = 100
    if len(mols) < batch_size:
        morgan = pd.DataFrame().join(fps, how='outer')
    else:
        morgans = []
        for start in trange(0, len(mols), batch_size, desc='Joining DataFrames'):
            morgan = pd.DataFrame().join(fps[start:start+batch_size], how='outer')
            morgans.append(morgan)
        morgan = pd.DataFrame().join(morgans, how='outer')
    morgan = morgan.fillna(0).T
    if features is not None:
        morgan = morgan[features]  # Retain only the specified features
    return morgan


def smiles_to_morgan_sim(smiles, ref_smiles, radius=5, features=None):
    mols = odorants.smiles_to_mol(smiles)
    ref_mols = odorants.smiles_to_mol(ref_smiles)
    morgan_sim = mol_to_morgan_sim(mols, ref_mols, radius=radius, features=features)
    return morgan_sim


def mol_to_morgan_sim(mols, ref_mols, radius=5, features=None):
    all_mols = pd.concat([pd.Series(mols), pd.Series(ref_mols)]).drop_duplicates()
    fps = all_mols.apply(lambda x: AllChem.GetMorganFingerprint(x, radius))
    fps = fps[~fps.index.duplicated()]
    morgan_sim = pd.DataFrame(index=mols.keys(), columns=ref_mols.keys())
    def dice(col):
        return np.array([DataStructs.DiceSimilarity(fps[col.name], fps[key]) for key in col.index])
    for ref in tqdm(ref_mols):
        morgan_sim[ref] = dice(morgan_sim[ref])
    return morgan_sim
    
    #print("%d similarity features for %d molecules" % morgan.shape[::-1])
    #return morgan


def mol_file_to_smiles(mol_file_path):
    # Takes a path to a .mol file and returns a SMILES string
    x = Chem.MolFromMolFile(mol_file_path)
    result = Chem.MolToSmiles(x, isomericSmiles=True)
    return result



from pyrfume.odorants import Compound, ChemicalOrder, Vendor, Molecule
from pyrfume.experiments import TriangleTest
from pyrfume.objects import Component, Mixture
from datetime import datetime
import quantities as pq


def get_water(substance_type: str = "mixtures"):
    vendor = Vendor("Test_Vendor", "")

    cid_H2O = 962
    cas_H2O = "7732-18-5"
    molecule_H2O = Molecule(cid_H2O, "Water", True)
    molecule_H2O.density = 	0.902 * pq.g / pq.mL
    molecule_H2O.molecular_weight = 	88.106 * pq.g / pq.mol
    molecule_H2O.cas = cas_H2O
    chemical_order_molecule_H2O = ChemicalOrder(molecule_H2O, vendor, "part 0", 0.5, None)
    compound_H2O = Compound(chemical_order_molecule_H2O, "TEST", datetime.now, datetime.now, True)
    component_H2O = Component(cid_H2O, "H2O", cas_H2O, 0.5, compound_H2O)
    mixture_H2O = Mixture(2, [component_H2O])
    descriptors = {
        cas_H2O : ["H2O unique descriptor", "common descriptor"],
        "dravnieks" : ["H2O dravnieks descriptor", "common dravnieks descriptor"],
        "sigma_ff" : ["H2O sigma_ff descriptor", "common sigma_ff descriptor"]
    }
    component_H2O.set_descriptors('unittest source', descriptors)

    if substance_type == "mixtures":
        return mixture_H2O
    elif substance_type == "components":
        return component_H2O
    elif substance_type == "compounds":
        return compound_H2O
    elif substance_type == "molecules":
        return molecule_H2O
    elif substance_type == "chemical_order":
        return chemical_order_molecule_H2O


def get_substances(substance_type: str = "mixtures"):
    vendor = Vendor("Test_Vendor", "")

    cid_C4H8O2 = 8857
    cas_C4H8O2 = "141-78-6"
    molecule_C4H8O2 = Molecule(cid_C4H8O2, "Ethyl acetate", True)
    molecule_C4H8O2.density = 	0.902 * pq.g / pq.mL
    molecule_C4H8O2.molecular_weight = 	88.106 * pq.g / pq.mol
    molecule_C4H8O2.vapor_pressure = 12.425 * pq.kPa
    molecule_C4H8O2.cas = cas_C4H8O2
    chemical_order_molecule_C4H8O2 = ChemicalOrder(molecule_C4H8O2, vendor, "part 0", 0.5, None)
    compound_C4H8O2 = Compound(chemical_order_molecule_C4H8O2, "TEST", datetime.now, datetime.now, False)
    component_C4H8O2 = Component(cid_C4H8O2, "C4H8O2", cas_C4H8O2, 0.5, compound_C4H8O2)
    mixture_C4H8O2 = Mixture(2, [component_C4H8O2])
    descriptors = {
        cas_C4H8O2 : ["C4H8O2 unique descriptor", "common descriptor"],
        "dravnieks" : ["C4H8O2 dravnieks descriptor", "common dravnieks descriptor"],
        "sigma_ff" : ["C4H8O2 sigma_ff descriptor", "common sigma_ff descriptor"]
    }
    component_C4H8O2.set_descriptors('unittest source', descriptors)

    cid_C2H6O = 702
    cas_C2H6O = "64-17-5"
    molecule_C2H6O = Molecule(cid_C2H6O, "Ethanol", True)
    molecule_C2H6O.density = 0.7893 * pq.g / pq.mL
    molecule_C2H6O.molecular_weight = 46.069 * pq.g / pq.mol
    molecule_C2H6O.vapor_pressure = 7.906 * pq.kPa
    molecule_C2H6O.cas = cas_C2H6O
    chemical_order_C2H6O = ChemicalOrder(molecule_C2H6O, vendor, "part 1", 0.5, None)
    compound_C2H6O = Compound(chemical_order_C2H6O, "TEST", datetime.now, datetime.now, False)
    component_C2H6O = Component(cid_C2H6O, "C2H6O", cas_C2H6O, 0.5, compound_C2H6O)
    mixture_C2H6O = Mixture(2, [component_C2H6O])
    descriptors = {
        cas_C2H6O : ["C2H6O unique descriptor", "common descriptor"],
        "dravnieks" : ["C2H6O dravnieks descriptor", "common dravnieks descriptor"],
        "sigma_ff" : ["C2H6O sigma_ff descriptor", "common sigma_ff descriptor"]
    }
    component_C2H6O.set_descriptors('unittest source', descriptors)

    cid_C4H8S = 1127
    cas_C4H8S = "110-01-0"
    molecule_C4H8S = Molecule(cid_C4H8S, "Tetrahydrothiophene", True)
    molecule_C4H8S.density = 0.997 * pq.g / pq.mL
    molecule_C4H8S.molecular_weight = 88.17 * pq.g / pq.mol
    molecule_C4H8S.vapor_pressure = 2.4 * pq.kPa
    chemical_order_C4H8S = ChemicalOrder(molecule_C4H8S, vendor, "part 2", 0.5, None)
    compound_C4H8S = Compound(chemical_order_C4H8S, "TEST", datetime.now, datetime.now, False)
    component_C4H8S = Component(cid_C4H8S, "C4H8S", cas_C4H8S, 0.5, compound_C4H8S)
    mixture_C4H8S = Mixture(2, [component_C4H8S])

    descriptors = {
        cas_C4H8S : ["C4H8S unique descriptor", "common descriptor"],
        "dravnieks" : ["C4H8S dravnieks descriptor", "common dravnieks descriptor"],
        "sigma_ff" : ["C4H8S sigma_ff descriptor", "common sigma_ff descriptor"]
    }
    component_C4H8S.set_descriptors('unittest source', descriptors)

    if substance_type == "mixtures":
        return (mixture_C4H8O2, mixture_C4H8S, mixture_C2H6O)
    elif substance_type == "components":
        return (component_C4H8O2, component_C4H8S, component_C2H6O)
    elif substance_type == "compounds":
        return (compound_C4H8O2, compound_C4H8S, compound_C2H6O)
    elif substance_type == "molecules":
        return (molecule_C4H8O2, molecule_C4H8S, molecule_C2H6O)
    elif substance_type == "chemical_order":
        return (chemical_order_molecule_C4H8O2, chemical_order_C4H8S, chemical_order_C2H6O)



""" Definitions of datajoint tables and functions for operating the datajoint schema."""

import datajoint as dj
from datajoint.user_tables import UserTable
import quantities as pq
from pyrfume import read_config, write_config, init_config
import inspect
import warnings
from typing import Dict

dj.errors._switch_adapted_types(True)
schema = None


def set_schema_name(new_schema_name: str) -> None:
    """ Set a new schema name into the config file. The data saved 
        into the old schema will be keeped on the server unless 
        `drop_schema` was called.

    Args:
        new_schema_name (str): The new name of the schema.
    """
    init_config(False)
    write_config("DATABASE", "schema_name", new_schema_name)


def init_schema() -> None:
    """ Get the name of schema from the config file 
        and initialize the schema with it.
    """

    schema_name = read_config("DATABASE", "schema_name")
    if schema_name == "":
        warnings.warn("Schema name was not set in pyrfume config.")
        return
    context = globals()
    global schema
    schema = dj.schema(schema_name, context)
    schema.context.update({'quantity_adapter': QuantityAdapter()})


def drop_schema(force=False) -> None:
    """Drop the datajoint schema.

    Args:
        force (bool, optional): Dropping it without a pop up confirmation. 
        Defaults to False.
    """

    global schema
    if not schema:
        init_schema()
    schema.drop(force=force)
    schema = None


def get_table(table_name: str) -> UserTable:
    """ Get a single datajoint table by table name.

    Args:
        table_name (str): The name of the datajoint table.

    Returns:
        UserTable: The datajoint table.
    """
    global schema
    if not schema:
        init_schema()

    result = None
    if table_name:
        try:
            cls = globals[table_name]
            if inspect.isclass(cls):
                result = schema(cls)
        except KeyError:
            print("Class does not exist.")

    return result


def get_tables() -> Dict[str, UserTable]:
    """ Get all datajoint tables in this module.

    Returns:
        Dict[str, UserTable]: The dict that contains all the table classes.
    """

    global schema
    if not schema:
        init_schema()

    result = {}
    for key, value in globals().items():
        if inspect.isclass(value) and value is not QuantityAdapter and value is not UserTable:
            result[key] = (schema(value))

    return result


class QuantityAdapter(dj.AttributeAdapter):
    """ The datajoint adapter class that puts and gets Python Quantity objects 
        to and from the datajoint database server.
    """
    attribute_type = 'float'

    def put(self, obj: pq.Quantity):
        assert isinstance(obj, pq.Quantity)
        obj = obj.rescale(pq.mL)
        return obj.item()

    def get(self, value: float):
        return value * pq.mL


class Molecule(dj.Manual):
    definition = '''
    smiles : varchar(25)
    ---
    inchi = NULL : varchar(128)
    inchikey = "" : varchar(256)
    pubchem_id = NULL : int
    name = "" : varchar(128)
    iupac = "" : varchar(128)
    '''


class Vendor(dj.Manual):
    definition = '''
    name = "" : varchar(256)
    ---
    '''


class Product(dj.Manual):
    definition = '''
    -> Vendor 
    catalog = 0 : int
    ---
    -> Molecule
    purity = "" : varchar(64)
    batch = "" : varchar(64)
    '''


class Compound(dj.Manual):
    definition = '''
    -> Product
    date_delivered : datetime
    location = "" : varchar(64)
    ---
    date_opened = NULL : datetime
    '''


class Solution(dj.Manual):
    definition = '''
    solution_id: int auto_increment
    ---
    diution = NULL : int
    concentration = NULL : float
    value = NULL : <quantity_adapter>
    mixing_data = NULL : date
    '''

    class Compounds(dj.Part):
        definition = '''
        -> Solution
        -> Compound
        ---
        '''


class Vessel(dj.Manual):
    definition = '''
    name = "" : varchar(64)
    height = 0 : float
    base_area = "" : varchar(64)
    ---
    '''


class Odorant(dj.Manual):
    definition = '''
    odorant_id: int auto_increment
    ---
    -> Vessel
    date_prepared = NULL : datetime
    '''

    class Solutions(dj.Part):
        definition = '''
        -> Odorant
        -> Solution
        ---
        '''


class Route(dj.Manual):
    definition = '''
    route_id: int auto_increment
    ---
    name = "" : varchar(64)
    '''


class Stimulus(dj.Manual):
    definition = '''
    stimulus_id: int auto_increment
    ---
    -> Route
    '''

    class Odorants(dj.Part):
        definition = '''
        -> Stimulus
        -> Odorant
        ---
        '''


class Subject(dj.Manual):
    definition = '''
    subject_id: int auto_increment
    ---
    age : tinyint
    gender : tinyint
    detail_info : varchar(512)
    '''


class Trial(dj.Manual):
    definition = '''
    trial_id: int auto_increment
    ---
    -> Stimulus
    -> Subject
    time : timestamp
    '''


class Site(dj.Manual):
    definition = '''
    site_id: int auto_increment
    ---
    name : varchar(64)
    kind : varchar(16)
    '''


class Investigator(dj.Manual):
    definition = '''
    investigator_id: int auto_increment
    ---
    first_name : varchar(64)
    last_name : varchar(64)
    -> Site
    '''


class Technician(dj.Manual):
    definition = '''
    technician_id: int auto_increment
    ---
    first_name : varchar(64)
    last_name : varchar(64)
    -> Investigator
    '''


class Publication(dj.Manual):
    definition = '''
    publication_id: int auto_increment
    ---
    name : varchar(512)
    kind : varchar(32)
    -> Investigator
    '''


class Report(dj.Manual):
    definition = '''
    report_id: int auto_increment
    ---
    title : varchar(512)
    year : smallint
    -> Publication
    doi : varchar(128)
    
    last_name : varchar(64)
    -> Investigator
    '''


class Design(dj.Manual):
    definition = '''
    design_id: int auto_increment
    ---
    name : varchar(64)
    '''


class Block(dj.Manual):
    definition = '''
    block_id: int auto_increment
    ---
    -> Technician
    -> Design
    '''

    class Trials(dj.Part):
        definition = '''
        -> Block
        -> Trial
        ---
        '''


class Experiment(dj.Manual):
    definition = '''
    experiment_id: int auto_increment
    ---
    -> Investigator
    '''

    class Blocks(dj.Part):
        definition = '''
        -> Experiment
        -> Block
        ---
        '''


class Summary(dj.Manual):
    definition = '''
    summary_id: int auto_increment
    ---
    -> Publication
    -> Design
    '''

    class Odorants(dj.Part):
        definition = '''
        -> Summary
        -> Odorant
        ---
        '''



"""
dravieks.txt produced by copying and pasting from the supplemental material of
"In search of the structure of human olfactory space", e.g. koulakov.pdf.
"""

import os

import pyrfume

DRAV_PATH = os.path.join(pyrfume.DATA_DIR, "dravnieks")


def get_data():
    cas_path = os.path.join(DRAV_PATH, "dravnieks_cas.txt")
    with open(cas_path, "r") as f:
        text = f.read()
    lines = text.split("\n")[1:]
    # A list of CAS numbers.
    cas = [line.split(" ")[1] for line in lines]

    descriptor_path = os.path.join(DRAV_PATH, "dravnieks_descriptors.txt")
    with open(descriptor_path, "r") as f:
        text = f.read()
    lines = text.split("\n")[1:]
    # A list of descriptors.
    descriptors = [" ".join(line.split(" ")[1:]) for line in lines]

    data_path = os.path.join(DRAV_PATH, "dravnieks_data.txt")
    with open(data_path, "r") as f:
        text = f.read()
    lines = text.split("\n")[:-1]
    data = {}
    for i, line in enumerate(lines):
        values = line.split("\t")
        data[cas[i]] = {descriptors[j]: float(value) for j, value in enumerate(values)}

    return cas, descriptors, data


if __name__ == "__main__":
    cas, descriptors, data = get_data()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyrfume', 'pyrfume.loaders', 'pyrfume.unit_test']

package_data = \
{'': ['*']}

install_requires = \
['black>=19.10b0',
 'dask[bag]>=2.17.0,<3.0.0',
 'datajoint>=0.12.6b1',
 'deap>=1.3.1',
 'eden-kernel>=0.3.1335',
 'flake8>=3.8.2',
 'ipython>=7.14.0',
 'ipywidgets>=7.5.1',
 'isort>=4.3.21',
 'matplotlib>=3.2.1',
 'mordred>=1.2.0',
 'numpy>=1.18.4',
 'pandas>=1.0.3',
 'plotly>=4.8.0',
 'pubchempy>=1.0.4',
 'quantities>=0.12.4',
 'requests>=2.20.0',
 'scikit-learn>=0.23.1',
 'scipy>=1.4.1',
 'sympy>=1.6',
 'toml>=0.10.2',
 'tqdm>=4.46.0']

setup_kwargs = {
    'name': 'pyrfume',
    'version': '0.17.3',
    'description': 'A validation library for human olfactory psychophysics research.',
    'long_description': '# Pyrfume\n\n![Pyrfume logo](https://avatars3.githubusercontent.com/u/34174393?s=200&v=4)\n\n#### `pyrfume` is a python library for olfactory psychophysics research. See "notebooks" for examples of use.\n[![Travis](https://travis-ci.org/pyrfume/pyrfume.svg?branch=master)](https://travis-ci.org/pyrfume/pyrfume) \n[![Coverage Status](https://coveralls.io/repos/github/pyrfume/pyrfume/badge.svg?branch=master)](https://coveralls.io/github/pyrfume/pyrfume?branch=master)\n\n### Examples:\n```\n# Load data for Snitz et al, 2013 (PLOS Computational Biology)\nimport pyrfume\nbehavior = pyrfume.load_data(\'snitz_2013/behavior.csv\')\nmolecules = pyrfume.load_data(\'snitz_2013/molecules.csv\')\n\n# Load data for Bushdid et al, 2014 (Science)\nimport pyrfume\nbehavior = pyrfume.load_data(\'bushdid_2014/behavior.csv\')\nmolecules = pyrfume.load_data(\'bushdid_2014/molecules.csv\')\nmixtures = pyrfume.load_data(\'bushdid_2014/behavior.csv\')\n```\n\n### [Website](http://pyrfume.org)\n\n### [Data Curation Status](http://status.pyrfume.org)\n\n### [Docs](https://pyrfume.readthedocs.io/)\n',
    'author': 'Rick Gerkin',
    'author_email': 'rgerkin@asu.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://pyrfume.org',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)



import json
import unittest
import quantities as pq
from quantities.units.velocity import speed_of_light
from pyrfume import save_data, set_data_path, get_data_path
from pyrfume.odorants import url_to_json, Solution, get_cid, \
                            get_cids, from_cids, cids_to_cas, \
                            cids_to_smiles, cactus, cactus_image, \
                            get_compound_summary, get_compound_odor, \
                            _parse_other_info, smiles_to_image, crop_image, \
                            all_odorants, all_sources, all_cids, all_smiles
                            

from .unittest_utils import get_substances

class OdorantsTestCase(unittest.TestCase):
    def setUp(self):
        self.compound_C4H8O2, self.compound_C4H8S, self.compound_C2H6O = get_substances("compounds")
        
        self.compounds = {
            self.compound_C4H8O2 : 100 * pq.mL,
            self.compound_C4H8S : 100 * pq.mL,
            self.compound_C2H6O : 100 * pq.mL
        }

    def test_solution(self):
        components = self.compounds
        solution = Solution(components, None)
        for key, value in self.compounds.items():
            self.assertTrue(key in solution.compounds.keys())
            self.assertTrue(value in solution.compounds.values())

        for compound in self.compounds.keys():
            self.assertTrue(compound in solution.compounds)

        components = {solution : 100 * pq.mL}
        solution = Solution(components, None)
        for compound in self.compounds.keys():
            self.assertTrue(compound in solution.compounds)
        
        molecule_C4H8O2 = self.compound_C4H8O2.chemical_order.molecule
        molecule_C4H8S = self.compound_C4H8S.chemical_order.molecule
        molecule_C2H6O = self.compound_C2H6O.chemical_order.molecule
        
        self.assertAlmostEqual(solution.molecules[molecule_C4H8O2].item(), 1.02376683, 3)
        self.assertAlmostEqual(solution.molecules[molecule_C4H8S].item(), 1.1307701, 3)
        self.assertAlmostEqual(solution.molecules[molecule_C2H6O].item(), 1.71329962, 3)

        self.assertAlmostEqual(solution.molarities[molecule_C4H8O2].item(), 0.01023767, 3)
        self.assertAlmostEqual(solution.molarities[molecule_C4H8S].item(), 0.0113077, 3)
        self.assertAlmostEqual(solution.molarities[molecule_C2H6O].item(), 0.017133, 3)

        self.assertAlmostEqual(solution.mole_fraction(molecule_C4H8O2), 0.2646872, 3)
        self.assertAlmostEqual(solution.mole_fraction(molecule_C4H8S), 0.2923520, 3)
        self.assertAlmostEqual(solution.mole_fraction(molecule_C2H6O), 0.442960, 3)

        self.assertAlmostEqual(solution.partial_pressure(molecule_C4H8O2).item(), 3.28873846, 3)
        self.assertAlmostEqual(solution.partial_pressure(molecule_C4H8S).item(), 0.70164502, 3)
        self.assertAlmostEqual(solution.partial_pressure(molecule_C2H6O).item(), 3.50204736, 3)

        self.assertAlmostEqual(solution.total_pressure.item(), 7492.4308428, 4)

        self.assertAlmostEqual(solution.vapor_concentration(molecule_C4H8O2), 0.032457325, 3)
        self.assertAlmostEqual(solution.vapor_concentration(molecule_C4H8S), 0.006924697, 3)
        self.assertAlmostEqual(solution.vapor_concentration(molecule_C2H6O), 0.034562520, 3)

        self.assertAlmostEqual(solution.molar_evaporation_rate(molecule_C4H8O2).item(), 0.00116778, 3)
        self.assertAlmostEqual(solution.molar_evaporation_rate(molecule_C4H8S).item(), 0.00023938, 3)
        self.assertAlmostEqual(solution.molar_evaporation_rate(molecule_C2H6O).item(), 0.00122993, 3)

    def test_molecular(self):
        molecule_C4H8O2_1, molecule_C4H8S_1, molecule_C2H6O_1 = get_substances("molecules")
        molecule_C4H8O2_2, molecule_C4H8S_2, molecule_C2H6O_2 = get_substances("molecules")
        self.assertTrue(molecule_C4H8O2_1 == molecule_C4H8O2_2)
        self.assertTrue(molecule_C4H8S_1 == molecule_C4H8S_2)
        self.assertTrue(molecule_C2H6O_1 == molecule_C2H6O_2)

        self.assertTrue(molecule_C4H8O2_1 > molecule_C4H8S_1)
        self.assertTrue(molecule_C4H8S_1 > molecule_C2H6O_1)
        self.assertTrue(molecule_C2H6O_1 < molecule_C4H8O2_1)

        self.assertIsInstance(repr(molecule_C4H8O2_1), str)
        self.assertIsInstance(repr(molecule_C4H8S_1), str)
        self.assertIsInstance(repr(molecule_C2H6O_1), str)

    def test_url_to_json(self):
        json_dict = url_to_json("https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/cid/962/synonyms/JSON")
        self.assertEqual(json_dict["InformationList"]["Information"][0]["Synonym"][0], "water")

    def test_get_cid(self):
        self.assertEqual(get_cid("64-17-5"), 702)
        self.assertEqual(get_cid("141-78-6"), 8857)
        self.assertEqual(get_cid("110-01-0"), 1127)

        cids = get_cids(["64-17-5",  "141-78-6", "110-01-0"])
        self.assertEqual(cids["64-17-5"], 702)
        self.assertEqual(cids["141-78-6"], 8857)
        self.assertEqual(cids["110-01-0"], 1127)

    def test_from_cids(self):
        results = from_cids([702, 8857, 1127])
        self.assertEqual(results[0]['IUPACName'], 'ethanol')
        self.assertEqual(results[1]['IUPACName'], 'ethyl acetate')
        self.assertEqual(results[2]['IUPACName'], 'thiolane')

    def test_cids_to_smiles(self):
        results = cids_to_smiles([702, 8857, 1127])
        self.assertEqual(results[702], 'CCO')
        self.assertEqual(results[8857], 'CCOC(=O)C')
        self.assertEqual(results[1127], 'C1CCSC1')

    def test_cids_to_cas(self):
        results = cids_to_cas([702, 8857, 1127])
        self.assertEqual(results[702][0], '64-17-5')
        self.assertEqual(results[8857][0], '141-78-6')
        self.assertEqual(results[1127][0], '110-01-0')

    def test_cactus(self):
        results = cactus("Ethyl acetate", "cas").split("\n")
        self.assertTrue("141-78-6" in results)

        results = cactus("Ethanol", "cas").split('\n')
        self.assertTrue("64-17-5" in results)

        results = cactus("Tetrahydrothiophene", "cas").split('\n')
        self.assertTrue("110-01-0" in results)

    def test_get_compound_summary(self):
        self.assertIsInstance(get_compound_summary(702, 'Physical Description'), dict)
        self.assertIsInstance(get_compound_summary(8857, 'Physical Description'), dict)
        self.assertIsInstance(get_compound_summary(1127, 'Physical Description'), dict)

    def test_get_compound_odor(self):
        self.assertIsInstance(get_compound_odor(702, False), list)
        self.assertIsInstance(get_compound_odor(8857, False), list)
        self.assertIsInstance(get_compound_odor(1127, False), list)
        self.assertIsInstance(get_compound_odor(702, True), list)
        self.assertIsInstance(get_compound_odor(8857, True), list)
        self.assertIsInstance(get_compound_odor(1127, True), list)

    def test__parse_other_info(self):
        self.assertTrue(len(_parse_other_info(None)) == 0)

        test_info_dict = {
            "String" : "value of String",
            "Value" : "Number 0",
            "Dict" : {"String" : "value of String in the sub-dict"}
        }
        result = _parse_other_info(test_info_dict)
        self.assertTrue('value of String' in result)
        self.assertTrue('Number 0' in result)
        self.assertTrue('value of String in the sub-dict' in result)

        test_info_list = [test_info_dict]
        result = _parse_other_info(test_info_list)
        self.assertTrue('value of String' in result)
        self.assertTrue('Number 0' in result)
        self.assertTrue('value of String in the sub-dict' in result)

    @unittest.skip("Failure expected. A fix is needed.")
    def test_cactus_image(self):
        cactus_image("CCC")

    @unittest.skip("Failure expected. The data is needed for this test case.")
    def test_all_items(self):
        from pandas import DataFrame
        #set_data_path('C:/icon/pyrfume')
        self.assertIsInstance(all_odorants(), DataFrame)
        self.assertIsInstance(all_sources(), DataFrame)
        self.assertIsInstance(all_cids(), list)
        self.assertIsInstance(all_smiles(), list)

if __name__ == '__main__':
    unittest.main()



import pandas as pd

import pyrfume
from pyrfume.features import smiles_to_mordred, smiles_to_morgan_sim
from pyrfume.odorants import all_smiles


def get_predicted_intensities():
    """Return the DREAM model predicted intensities using Mordred (not Dragon) features"""
    path = "physicochemical/cids-names-smiles-mordredpredintensities.csv"
    df = pyrfume.load_data(path)
    return df["Intensity"]


def get_predicted_odorless():
    """Return a pandas Series that is True for molecules predicted to have an odor
    and False for those predicted to be odorless"""
    path = "odorants/predicted_odorless.csv"
    df = pyrfume.load_data(path)
    return df["PredictedOdorless"]


def load_dream_model():
    """Load the open source DREAM model created in `open-source-dream.ipynb`"""
    path = "keller_2017/open-source-dream.pkl"
    model, use_features, descriptors, imputer = pyrfume.load_data(path)
    # model: The model to use for prediction
    # features: The features used for each column
    # descriptors: The descriptors it will predict for each output
    # imputer: The missing data imputer that should be used on any Morgan features
    #          before joining with other features and passing it to the model
    return model, use_features, descriptors, imputer


def smiles_to_features(smiles, use_features, imputer):
    """Turn SMILES into model input"""
    # Compute Mordred features for the new molecules
    mordred = smiles_to_mordred(smiles)
    # Impute any missing values based on the missing data imputer
    print(mordred.shape)
    mordred[:] = imputer.transform(mordred)
    # Compute Morgan features for the new molecules
    sim_smiles = list(set(all_smiles()))
    morgan_sim = smiles_to_morgan_sim(smiles, sim_smiles)
    # Combine Mordred (after imputation) and Morgan features into one dataframe
    features = mordred.join(morgan_sim)
    # Add the `Dilution` column
    features["Dilution"] = -3
    # Restrict to only those features used in the model, and in the same order
    features = features[use_features]
    # Make sure the list of features used for prediction is identical to that used for training
    # i.e. make sure the line above did what was intended
    assert list(features) == use_features
    return features


def predict(model, features, descriptors):
    # Predict perceptual descriptors for the new molecules based on their physicochemical features
    predictions = model.predict(features).round(2)
    # Make a dataframe for these predictions
    predictions = pd.DataFrame(index=features.index, columns=descriptors, data=predictions)
    return predictions



"""Convenience functions for working with DataJoint"""

import re
from importlib import import_module
from inspect import isclass
from typing import Any, ForwardRef, _GenericAlias
import datajoint as dj
from .dbtables import QuantityAdapter
dj.errors._switch_adapted_types(True)

QUANTITY_ADAPTER = None

def schematize(cls, schema: dj.schema):
    """Take a Python class and build a Datajoint table from it.

    Params:
        cls: The class to convert into a DataJoint table
        schema: The schema in which to add the table

    Returns:
        cls: The schematized class (now for use with DataJoint)
    """
    cls = type(cls.__name__, (dj.Manual, cls, object), {})
    cls = set_dj_definition(cls)
    global QUANTITY_ADAPTER
    if QUANTITY_ADAPTER:
        schema.context.update({'QUANTITY_ADAPTER': QUANTITY_ADAPTER})
    cls = schema(cls)
    return cls

def create_quantity_adapter() -> None:
    """ Create an datajoint adapter class, `QuantityAdapter`, that puts and gets 
        Python Quantity objects to and from the datajoint database server.
        The adapter will be assigned to the global variable `QUANTITY_ADAPTER`
        in this module.
    """

    global QUANTITY_ADAPTER
    QUANTITY_ADAPTER = QuantityAdapter()


def handle_dict(cls, _type_map: dict, attr: Any, type_hint: _GenericAlias):
    """Using master-part relation to store a dict. It is assumed that 
        the type of keys have corresponding tables in the database.

        It is assumed that values of the dict are:
        primitive type which is in `_type_map`
        OR
        `quantities.Quantity` instance.

    Args:
        _type_map (dict): A map that maps type hint to data type that accepted by datajoint.
        attr (Any): Variable name of the dict.
        type_hint (typing._GenericAlias): Required to be a type hint like `Dict[TypeA, int]`.
                                            A type hint of `dict` will cause an exception.

    Returns:
        type: `cls` that contains a part class for the keys of the dict.
    """

    # For example, components: Dict[ClassA, int] = {a: 1, b: 2}
    # key_cls_name will be "ClassA"
    # part_cls_name will be "Component", 
    # note that the "s" at the end of the dict name will be removed.

    part_cls_name = attr[0].upper() + attr[1:]
    part_cls_name = part_cls_name[:-1] if part_cls_name[-1] == 's' else part_cls_name

    key_type = type_hint.__args__[0]
    value_type = type_hint.__args__[1]

    key_cls_name = key_type.__forward_arg__ if isinstance(key_type, ForwardRef) else key_type.__name__
    value_type = value_type.__forward_arg__ if isinstance(value_type, ForwardRef) else value_type.__name__

    if value_type == 'Quantity':
        if not QUANTITY_ADAPTER:
            create_quantity_adapter()
        value_type = '<QUANTITY_ADAPTER>'
    else:
        assert value_type in _type_map
        value_type = _type_map[value_type]

    part_cls = type(
        part_cls_name,
        (dj.Part, object),
        {
            "definition": "\n-> %s\n-> %s\n---\nvalue = NULL : %s"
            % (cls.__name__, key_cls_name, value_type)
        }
    )
    cls_dict = dict(vars(cls))
    cls_dict[part_cls_name] = part_cls
    cls = type(cls.__name__, tuple(cls.__bases__), {part_cls_name: part_cls})
    return cls

def set_dj_definition(cls, type_map: dict = None) -> None:
    """Set the definition property of a class by inspecting its attributes.

    Params:
        cls: The class whose definition attribute should be set
        type_map: Optional additional type mappings
    """
    # A mapping between python types and DataJoint types
    _type_map = {
        "int": "int", 
        "str": "varchar(256)", 
        "float": "float",
        "Quantity": "float",
        "datetime": "datetime", 
        "datetime.datetime": "datetime", 
        "bool": "tinyint",
        "list": "longblob",
    }
    # A list of python types which have no DataJoint
    # equivalent and so are unsupported
    unsupported = [dict]
    if type_map:
        _type_map.update(type_map)
    dj_def = "%s_id: int auto_increment\n---\n" % cls.__name__.lower()
    cls_items = cls.__annotations__.items()
    for attr, type_hint in cls_items:
        if type_hint in unsupported:
            continue
        name = getattr(type_hint, "__name__", type_hint)
        default = getattr(cls, attr)            
        
        if isinstance(default, str):
            default = '"%s"' % default
        elif isinstance(default, bool):
            default = int(default)
        else:
            default = "NULL"

        if getattr(type_hint, '_name', "") == 'Dict':
            cls = handle_dict(cls, _type_map, attr, type_hint)
            continue
        elif name in _type_map:
            dj_def += "%s = %s : %s\n" % (attr, default, _type_map[name])
        else:
            dj_def += "-> %s\n" % name
    cls.definition = dj_def
    return cls


def import_classes(module_name: str, match: str = None) -> dict:
    """Import all classes from the named module'

    Params:
        module_name (str): Name of the module (e.g. 'pyrfume.odorants')
        match (str): Optional regex string for class names to match.

    Returns:
        dict: A dictionary of full class names and the classes themselves"""
    classes = {}
    # Import the module and iterate through its attributes
    module = import_module(module_name)
    for attr_name in dir(module):
        if match and not re.search(match, attr_name):
            continue
        attr = getattr(module, attr_name)
        if isclass(attr):
            # Add the class to this package's variables
            classes[attr_name] = attr
    return classes


# Demonstration of setting the definition.
# A demonstration of adding to the schema requires an active schema.
if __name__ == "__main__":

    class Stuff:
        """A regular Python class that does not use DataJoint.
        All of its attributes (that you want to use in DataJoint)
        must have type hints."""

        other: "OtherStuff" = None

        x: int = 3
        """type: int"""

        y: str = "something"
        """type: str"""

    # Set DataJoint definitions in each class based on inspection
    # of class attributes.
    set_dj_definition(Stuff)

    # Schematize each of these classes (will include foreign keys)
    from pyrfume.odorants import Molecule, Vendor, ChemicalOrder

    for cls in [Molecule, Vendor, ChemicalOrder]:
        locals()[cls.__name__] = djt.schematize(cls, schema)

    # May require a Jupyter notebook or other canvas
    dj.ERD(schema).draw()
